##
# This file is part of TALER
# (C) 2014, 2015, 2016, 2020 Taler Systems SA
#
# TALER is free software; you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation; either version 3, or
# (at your option) any later version. TALER is distributed in the
# hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
#  You should have received a copy of the GNU General Public
# License along with TALER; see the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>
#
#  @author Marcello Stanisci
#  @author Florian Dold

from functools import wraps
import math
import json
import logging
import hashlib
import random
import re
import time
import base64
import uuid
from urllib.parse import urlparse, parse_qsl
import django.contrib.auth
import django.contrib.auth.views
import django.contrib.auth.forms
from django.db import transaction
from django import forms
from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_POST, require_GET
from django.utils.translation import gettext
from django.views.decorators.http import require_http_methods
from django.urls import reverse, get_script_prefix
from django.contrib.auth.models import User
from django.db.models import Q
from django.http import JsonResponse, HttpResponse
from django.shortcuts import render, redirect
from django.core.exceptions import ObjectDoesNotExist
from datetime import datetime
from .models import BankAccount, BankTransaction, TalerWithdrawOperation
from taler.util.amount import Amount, SignedAmount
from taler.util.payto import PaytoParse
from taler.util.taler_error_codes import ErrorCode
from http import HTTPStatus

import qrcode
import qrcode.image.svg
import lxml
from .schemas import (
    HistoryParams,
    URLParamValidationError,
    AddIncomingData,
    JSONFieldException,
    InvalidSession,
    WithdrawHeadless,
    WithdrawHeadlessUri,
)

LOGGER = logging.getLogger(__name__)

##
# Constant value for the biggest number the bank handles.
# This value is just equal to the biggest number that JavaScript
# can handle (because of the wallet).
UINT64_MAX = (2 ** 64) - 1

##
# Decorator function that authenticates requests by fetching
# the credentials over the HTTP requests headers.
#
# @param view_func function that will be called after the
#        authentication, and that will usually serve the requested
#        endpoint.
# @return FIXME.
def login_via_headers(view_func):
    def _decorator(request, *args, **kwargs):
        user_account = basic_auth(request)
        if not user_account:
            raise LoginFailed("authentication failed")
        return view_func(request, user_account, *args, **kwargs)

    return wraps(view_func)(_decorator)


def allow_origin_star(view_func):
    def _decorator(request, *args, **kwargs):
        response = view_func(request, *args, **kwargs)
        response["Access-Control-Allow-Origin"] = "*"
        return response

    return wraps(view_func)(_decorator)

##
# Exception raised upon failing login.
#
class LoginFailed(Exception):
    def __init__(self, msg):
        super(LoginFailed, self).__init__(msg)
        self.hint = "Wrong password given"
        self.http_status_code = HTTPStatus.UNAUTHORIZED
        self.taler_error_code = ErrorCode.BANK_LOGIN_FAILED


class InvalidInputData(Exception):
    def __init__(self, msg):
        super(InvalidInputData, self).__init__(msg)
        self.hint = msg  # should mention the picked username
        self.http_status_code = HTTPStatus.BAD_REQUEST
        self.taler_error_code = ErrorCode.BANK_SOFT_EXCEPTION


class UsernameUnavailable(Exception):
    def __init__(self, msg):
        super(UsernameUnavailable, self).__init__(msg)
        self.hint = msg  # should mention the picked username
        self.http_status_code = HTTPStatus.NOT_ACCEPTABLE
        self.taler_error_code = ErrorCode.BANK_SOFT_EXCEPTION


##
# Exception raised when the public history from
# a ordinary user account is tried to be accessed.
class PrivateAccountException(Exception):
    def __init__(self, msg):
        super(PrivateAccountException, self).__init__(msg)
        self.hint = "Cannot show history from private persons accounts"
        self.http_status_code = HTTPStatus.FORBIDDEN


##
# Exception raised when some financial operation goes
# beyond the limit threshold.
class DebitLimitException(Exception):
    def __init__(self, msg):
        super(DebitLimitException, self).__init__(msg)
        self.hint = "Payment aborted for insufficient credit"
        self.http_status_code = HTTPStatus.FORBIDDEN
        self.taler_error_code = ErrorCode.BANK_UNALLOWED_DEBIT


##
# Exception raised when some financial operation is
# attempted and both parties are the same account number.
#
class SameAccountException(Exception):
    def __init__(self, msg):
        super(SameAccountException, self).__init__(msg)
        self.hint = "Cannot send payment to oneself."
        self.http_status_code = HTTPStatus.BAD_REQUEST
        self.taler_error_code = ErrorCode.BANK_SAME_ACCOUNT


class UnhandledException(Exception):
    def __init__(self, msg="Unhandled exception happened!"):
        super(UnhandledException, self).__init__(msg)
        self.hint = msg
        self.http_status_code = HTTPStatus.INTERNAL_SERVER_ERROR
        self.taler_error_code = ErrorCode.BANK_UNMANAGED_EXCEPTION


##
# The authentication for users to log in the bank.
#
class TalerAuthenticationForm(django.contrib.auth.forms.AuthenticationForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["username"].widget.attrs["autofocus"] = True
        self.fields["username"].widget.attrs["placeholder"] = "Username"
        self.fields["password"].widget.attrs["placeholder"] = "Password"


##
# Return a empty response.  Used in "/favicon.ico" requests.
#
def ignore(request):
    del request
    return HttpResponse()


##
# Decode body, when it is expected to be UTF-8.
#
# @param request the HTTP request being served.
# @return the body as string.
def decode_body(request):
    return request.body.decode("utf-8")


##
# Get a flag from the session and clear it.
#
# @param request the HTTP request being served.
# @param name name of the session value that should be retrieved.
# @return the value, if found; otherwise False.
def get_session_flag(request, name):
    if name in request.session:
        ret = request.session[name]
        del request.session[name]
        return ret
    return False


##
# A session "hint" is a tuple indicating whether the
# message is for a failure or a success, and containing
# the message itself.
#
# @param request the HTTP request being served.
# @param name hint name
# @return the hint (a "null" one if none was found)
def get_session_hint(request):
    ret = True, ""
    if "hint" in request.session:
        ret = request.session["hint"]
        del request.session["hint"]
    return ret


def set_session_hint(request, success, hint):
    if "hint" in request.session:
        LOGGER.warning("Overriding a non consumed hint")
        del request.session["hint"]
    request.session["hint"] = success, hint


##
# Build the list containing all the predefined accounts; the
# list contains, for example, the exchange, the bank itself, and
# all the public accounts (like GNUnet / Tor / FSF / ..)
def predefined_accounts_list():
    account = 2
    ret = []
    for i in settings.TALER_PREDEFINED_ACCOUNTS[1:]:
        ret.append((account, "%s (#%d)" % (i, account)))
        account += 1
    return ret


##
# Thanks to [1], this class provides a dropdown menu that
# can be used within a <select> element, in a <form>.
# [1] https://stackoverflow.com/questions/24783275/django-form-with-choices-but-also-with-freetext-option
class InputDatalist(forms.TextInput):

    ##
    # Constructor function.
    #
    # @param self the object itself.
    # @param datalist a list of admitted values.
    # @param name the name of the value that will be sent
    #        along the POST.
    # @param args positional arguments
    # @param kwargs keyword arguments
    # @return the object
    def __init__(self, datalist, name, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._name = name
        self._datalist = datalist()
        self.attrs.update({"list": "%slist" % name, "pattern": "[1-9]+[0-9]*"})

    ##
    # Method that produces the final HTML from the object itself.
    #
    # @param self the object itself
    # @param name the name of the value that will be sent
    #        along the POST
    # @param value value to be sent along the @a name.
    # @param attrs a dict indicating which HTML attribtues should
    #        be defined in the rendered element.
    # @param renderer render engine (left as None, typically); it
    #        is a class that respects the low-level render API from
    #        Django, see [2]
    # [2] https://docs.djangoproject.com/en/2.1/ref/forms/renderers/#low-level-widget-render-api
    def render(self, name, value, attrs=None, renderer=None):
        html = super().render(name, value, attrs=attrs, renderer=renderer)
        datalist = '<datalist id="%slist">' % self._name
        for dl_value, dl_text in self._datalist:
            datalist += '<option value="%s">%s</option>' % (dl_value, dl_text)
        datalist += "</datalist>"
        return html + datalist


class PaytoTransferForm(forms.Form):
    address = forms.CharField()


@login_required
def wiretransfer_form(request):
    if request.method == "GET":
        is_success, hint = get_session_hint(request)
        context = dict(
            currency=request.user.bankaccount.balance.amount.currency,
            is_success=is_success,
            hint=hint,
        )
        return render(request, "wiretransfer.html", context)

    # A payment was submitted. 
    try:
        amount = Amount.parse(
            "{}:{}".format(request.POST.get("currency"), request.POST.get("amount")
        ))
    except Exception:
        set_session_hint(request, success=False, hint="Wrong amount specified.")
        return redirect("wiretransfer-form")
    try:
        receiver_user = User.objects.get(username=request.POST.get("receiver"))
    except User.DoesNotExist:
        set_session_hint(request, success=False, hint="Money receiver was not found")
        return redirect("wiretransfer-form")

    try:
        wire_transfer(
            amount,
            request.user.bankaccount,
            receiver_user.bankaccount,
            request.POST.get("subject", "not given")
        )
    except Exception as exception:
        hint = str(exception)
        if hasattr(exception, "hint"):
            hint = exception.hint
        set_session_hint(request, success=False, hint=hint)
        return redirect("wiretransfer-form")

    set_session_hint(request, success=True, hint=gettext("Wire transfer successful!"))
    return redirect("profile")

@login_required
def payto_form(request):
    is_success, hint = get_session_hint(request)
    context = dict(
        currency=request.user.bankaccount.balance.amount.currency,
        is_success=is_success,
        hint=hint,
    )
    return render(request, "payto_wiretransfer.html", context)


##
# This method serves the profile page, which is the main
# page where the user interacts with the bank, and also the
# page that the user gets after a successful login.  It has
# to handle the following cases: (1) the user requests the
# profile page after haing POSTed a wire transfer request.
# (2) The user requests the page after having withdrawn coins,
# that means that a wire transfer has been issued to the exchange.
# In this latter case, the method has to notify the wallet about
# the operation outcome.  (3) Ordinary GET case, where the
# straight page should be returned.
#
# @param request Django-specific HTTP request object.
# @return Django-specific HTTP response object.
@login_required
def profile_page(request):
    is_success, hint = get_session_hint(request)
    context = dict(
        name=request.user.username,
        balance=request.user.bankaccount.balance,
        is_success=is_success,
        hint=hint,
        precision=settings.TALER_DIGITS,
        currency=request.user.bankaccount.balance.amount.currency,
        account_no=request.user.bankaccount.account_no,
        history=extract_history(request.user.bankaccount, -1 * (UINT64_MAX / 2 / 2)),
    )
    if settings.TALER_SUGGESTED_EXCHANGE:
        context["suggested_exchange"] = settings.TALER_SUGGESTED_EXCHANGE

    response = render(request, "profile_page.html", context)
    if "just_withdrawn" in request.session:
        del request.session["just_withdrawn"]
        response["Taler"] = "taler://notify-reserve/"
        response.status_code = HTTPStatus.ACCEPTED
    return response


@login_required
@require_POST
def payto_transfer(request):
    data = PaytoTransferForm(request.POST)
    if not data.is_valid():
        set_session_hint(request, success=False, hint=gettext("Bad form submitted!"))
        return redirect("profile")

    parsed_address = PaytoParse(data.cleaned_data.get("address"))

    try:
        receiver_user = User.objects.get(username=parsed_address.target)
    except User.DoesNotExist:
        set_session_hint(request, success=False, hint="Money receiver was not found")
        return redirect("payto-form")

    wire_transfer(
        parsed_address.amount,
        BankAccount.objects.get(user=request.user),
        BankAccount.objects.get(user=receiver_user),
        parsed_address.message,
    )
    set_session_hint(request, success=True, hint=gettext("Wire transfer successful!"))
    return redirect("profile")


##
# Helper function that hashes its input.  Usually
# used to hash the response to the math CAPTCHA.
#
# @param answer the plain text answer to hash.
# @return the hashed version of @a answer.
def hash_answer(answer):
    hasher = hashlib.new("sha1")
    hasher.update(settings.SECRET_KEY.encode("utf-8"))
    hasher.update(answer.encode("utf-8"))
    return hasher.hexdigest()


##
# Helper function that makes CAPTCHA's question and
# answer pair.
#
# @return the question and (hashed) answer pair.
def make_question():
    num1 = random.randint(1, 10)
    operand = random.choice(("*", "+", "-"))
    num2 = random.randint(1, 10)
    if operand == "*":
        answer = str(num1 * num2)
    elif operand == "-":
        # ensure result is positive
        num1, num2 = max(num1, num2), min(num1, num2)
        answer = str(num1 - num2)
    else:
        answer = str(num1 + num2)
    question = "{} {} {}".format(num1, operand, num2)
    return question, hash_answer(answer)


def get_acct_from_payto(uri_str: str) -> str:
    wire_uri = urlparse(uri_str)
    if wire_uri.scheme != "payto":
        raise Exception("Bad Payto URI: '%s'" % uri_str)
    return wire_uri.path.split("/")[-1]


def get_subject_from_payto(uri_str: str) -> str:
    wire_uri = urlparse(uri_str)
    if wire_uri.scheme != "payto":
        raise Exception("Bad Payto URI: '%s'" % uri_str)
    params = parse_qs(wire_uri.query)
    subject = params.get("subject")
    if not subject:
        raise Exception("Subject not found in Payto URI: '%s'" % uri_str)
    return subject


##
# Class representing the registration form.
class UserReg(forms.Form):
    username = forms.CharField()
    password = forms.CharField(widget=forms.PasswordInput())


def internal_register(username, password):
    if not settings.ALLOW_REGISTRATIONS:
        raise Exception("registration not allowed!")

    if User.objects.filter(username=username).exists():
        raise UsernameUnavailable(f"Username {username} is unavailable")

    # Registration goes through.
    with transaction.atomic():
        bank_internal_account = BankAccount.objects.get(account_no=1)

        user = User.objects.create_user(username=username, password=password)
        user_account = BankAccount(user=user)
        user_account.save()

        # Give the user their joining bonus
        wire_transfer(
            Amount(settings.TALER_CURRENCY, 100, 0),
            bank_internal_account,
            user_account,
            "Joining bonus",
        )

    return user


@require_POST
@csrf_exempt
def register_headless(request):
    """
    This method serves the request for programmatically
    registering a user.
    """
    if not settings.ALLOW_REGISTRATIONS:
        return JsonResponse(
            dict(error="registrations are not allowed"), status=HTTPStatus.FORBIDDEN
        )
    username = expect_json_body_str(request, "username")
    password = expect_json_body_str(request, "password")
    try:
        internal_register(username, password)
    except UsernameUnavailable:
        return JsonResponse(
            dict(hint="username unavailable"), status=HTTPStatus.CONFLICT
        )
    except InvalidInputData:
        return HttpResponse(status=HTTPStatus.NOT_ACCEPTABLE)  # WTF? BAD REQUEST?

    return HttpResponse(status=HTTPStatus.OK)


def register(request):
    """
    This method serves the request for registering a user.
    If successful, it redirects the user to their profile page;
    otherwise it will show again the same form (currently, without
    displaying _any_ error/warning message.)
    """
    if not settings.ALLOW_REGISTRATIONS:
        # FIXME: shouldn't be JSON!
        return JsonResponse(
            dict(error="registrations are not allowed"), status=HTTPStatus.FORBIDDEN
        )
    if request.method != "POST":
        return render(request, "register.html")

    # Process POST.

    try:
        input_data = UserReg(request.POST)
        if not input_data.is_valid():
            msg = "Wrong field(s): %s." % ", ".join(input_data.errors.keys())
            raise InvalidInputData(msg)
        username = input_data.cleaned_data["username"]
        password = input_data.cleaned_data["password"]
        user = internal_register(username, password)
    except UsernameUnavailable as e:
        return render(request, "register.html", {"not_available": True})
    except InvalidInputData as e:
        return render(
            request,
            "register.html",
            {
                "wrong": True,
                "hint": gettext("Username or password were incorrect!"),
            },
        )

    except DebitLimitException as e:
        return render(
            request,
            "register.html",
            {"wrong": True, "hint": "Out of business, cannot admit new customers."},
        )

    set_session_hint(request, success=True, hint=gettext("Registration successful!"))

    django.contrib.auth.login(request, user)
    return redirect("profile")


##
# Logs the user out, redirecting it to the bank's homepage.
#
# @param request Django-specific HTTP request object.
# @return Django-specific HTTP response object.
def logout_view(request):
    django.contrib.auth.logout(request)
    return redirect("index")


@require_GET
def config_view(request):
    """
    Config query of the taler bank access api
    """
    return JsonResponse(
        dict(
            version="0:0:0", currency=settings.TALER_CURRENCY, name="taler-bank-access"
        ),
        status=HTTPStatus.OK,
    )


@require_GET
@allow_origin_star
def api_config(request):
    """
    Config query of the taler bank integration api
    """
    return JsonResponse(
        dict(
            version="0:0:0",
            currency=settings.TALER_CURRENCY,
            name="taler-bank-integration",
        ),
        status=HTTPStatus.OK,
    )


def extract_history(account, delta, start=None):
    history = []
    qs = query_history(account, "both", delta, start)
    for item in qs:
        if item.credit_account == account:
            counterpart = item.debit_account
            sign = ""
        else:
            counterpart = item.credit_account
            sign = "-"
        entry = dict(
            row_id=item.id,
            cancelled=item.cancelled,
            sign=sign,
            amount=item.amount.stringify(settings.TALER_DIGITS, pretty=True),
            counterpart=counterpart.account_no,
            counterpart_username=counterpart.user.username,
            subject=item.subject,
            date=item.date.strftime("%d/%m/%y %H:%M %z"),
        )
        history.append(entry)
    return history


##
# Serve the page showing histories from publicly visible accounts.
#
# @param request Django-specific HTTP request object.
# @param name name of the public account to show.
# @param page given that public histories are paginated, this
#        value is the page number to display in the response.
# @return Django-specific HTTP response object.
def serve_public_accounts(request, name=None, page=None):
    try:
        page = abs(int(page))
        if page == 0:
            raise Exception
    except Exception:
        page = 1

    if not name:
        name = settings.TALER_PREDEFINED_ACCOUNTS[0]
    user = User.objects.get(username=name)
    if not user.bankaccount.is_public:
        raise PrivateAccountException(
            "Can't display public history for private account"
        )

    # How many records does a user have.
    num_records = query_history(
        user.bankaccount,
        "both",
        # Note: the parameter below is used for slicing arrays
        # and django/python is not allowing slicing with big numbers.
        UINT64_MAX / 2 / 2,
        0,
    ).count()
    DELTA = 30
    # '//' operator is NO floating point.
    num_pages = max(num_records // DELTA, 1)

    public_accounts = BankAccount.objects.filter(is_public=True)

    # Retrieve DELTA records younger than 'start_row' (included).
    history = extract_history(user.bankaccount, DELTA * page, 0)[
        DELTA * (page - 1) : (DELTA * page)
    ]

    pages = list(range(1, num_pages + 1))

    is_success, hint = get_session_hint(request)
    context = dict(
        is_success=is_success,
        hint=hint,
        current_page=page,
        back=page - 1 if page > 1 else None,
        forth=page + 1 if page < num_pages else None,
        public_accounts=public_accounts,
        selected_account=dict(
            name=name,
            number=user.bankaccount.account_no,
            history=history,
        ),
        pages=pages,
    )
    return render(request, "public_accounts.html", context)


##
# Build the DB query switch based on the "direction" history
# argument given by the user.
#
# @param bank_account bank account of the user requesting history.
# @param direction the "direction" URL parameter given by the user.
#        Note: this values got sanity-checked before this function
#        is called.
def direction_switch(bank_account, direction):
    direction_switch = {
        "both": (Q(debit_account=bank_account) | Q(credit_account=bank_account)),
        "credit": Q(credit_account=bank_account),
        "debit": Q(debit_account=bank_account),
        "cancel+": (Q(credit_account=bank_account) & Q(cancelled=True)),
        "cancel-": (Q(debit_account=bank_account) & Q(cancelled=True)),
    }
    return direction_switch.get(direction)


##
# Main routine querying for histories.
#
# @param bank_account the bank account object whose
#        history is being extracted.
# @param direction takes the following three values,
#        * debit: only entries where the querying user has _paid_
#                 will be returned.
#        * credit: only entries where the querying user _got_
#                  paid will be returned.
#        * both: both of the cases above will be returned.
#        * cancel+: only entries where the querying user cancelled
#                   the _receiving_ of money will be returned.
#        * cancel-: only entries where the querying user cancelled
#                   the _paying_ of money will be returned.
# @param delta how many history entries will be contained in the
#        array.
# @param start any history will be searched starting from this
#        value (which is a row ID), and going to the past or to
#        the future (depending on the user choice).  However, this
#        value itself will not be included in the history.
# @param sign this value ("+"/"-") determines whether the history
#        entries will be younger / older than @a start.
# @param ordering "descending" or anything else (for "ascending").
def query_history(bank_account, direction, delta, start):
    if start is None:
        if delta > 0:
            start = -1
        else:
            start = UINT64_MAX
    if delta < 0:
        sign_filter = Q(id__lt=start)
    else:
        sign_filter = Q(id__gt=start)

    qs = BankTransaction.objects.filter(
        direction_switch(bank_account, direction), sign_filter
    )
    order = "id" if (delta > 0) else "-id"
    return qs.order_by(order)[: abs(delta)]


##
# Build response object for /history.
#
# @param qs the query set for a history request.
# @param cancelled controls whether we omit/show
#        cancelled transactions.
# @param user_account bank account of the user who
#        asked for the history.
# @return the history object as specified in the
#         API reference documentation.
def build_history_response(qs, cancelled, user_account):
    history = []
    for entry in qs:
        counterpart = entry.credit_account.account_no
        sign_ = "-"
        if entry.cancelled and cancelled == "omit":
            continue
        if entry.credit_account.account_no == user_account.bankaccount.account_no:
            counterpart = entry.debit_account.account_no
            sign_ = "+"
        cancel = "cancel" if entry.cancelled else ""
        sign_ = cancel + sign_
        history.append(
            dict(
                counterpart=counterpart,
                amount=entry.amount.stringify(),
                sign=sign_,
                wt_subject=entry.subject,
                row_id=entry.id,
                date=dict(t_ms=int(entry.date.timestamp()) * 1000),
            )
        )
    return history


##
# Serve a request of /history.
#
# @param request Django-specific HTTP request.
# @param user_account the account whose history should be gotten.
# @return Django-specific HTTP response object.
@require_GET
@login_via_headers
def serve_history(request, user_account):
    args = HistoryParams(request.GET.dict())
    qs = query_history(
        user_account.bankaccount,
        args.get("direction"),
        args.get("delta"),
        args.get("start", None),
    )

    history = build_history_response(qs, args.get("cancelled", "show"), user_account)

    return JsonResponse(dict(data=history), status=HTTPStatus.OK)


def expect_json_body_str(request, param_name):
    body = json.loads(request.body)  # FIXME: cache!
    val = body.get(param_name)
    if not isinstance(val, str):
        # FIXME: throw right exception to be handled by middleware
        raise Exception(f"expected string for {param_name}")
    return val


def expect_json_body_amount(request, param_name):
    body = json.loads(request.body)  # FIXME: cache!
    val = body.get(param_name)
    if not isinstance(val, str):
        # FIXME: throw right exception to be handled by middleware
        raise Exception(f"expected string for {param_name}")
    return Amount.parse(val)


def expect_param_str(request, param_name):
    val = request.GET[param_name]
    if not isinstance(val, str):
        # FIXME: throw right exception to be handled by middleware
        raise Exception(f"expected string for {param_name}")
    return val


def expect_param_amount(request, param_name):
    val = request.GET[param_name]
    if not isinstance(val, str):
        # FIXME: throw right exception to be handled by middleware
        raise Exception(f"expected string for {param_name}")
    return Amount.parse(val)


@require_GET
def twg_base(request, acct_id):
    """
    This endpoint is used by the exchange test cases to
    check if the account is up, should not normally be used
    for anything else.
    """
    return JsonResponse(dict(), status=HTTPStatus.OK)


@require_GET
def twg_config(request, acct_id):
    """
    This endpoint is used by the exchange test cases to
    check if the account is up, should not normally be used
    for anything else.
    """
    return JsonResponse(
        dict(
            version="0:0:0",
            name="taler-wire-gateway",
            currency=settings.TALER_CURRENCY,
        ),
        status=HTTPStatus.OK,
    )


@csrf_exempt
@require_POST
@login_via_headers
def twg_add_incoming(request, user_account, acct_id):
    """
    Transfer money from a user's bank account to the exchange
    for testing purposes.
    """
    exchange_account = user_account.bankaccount

    if acct_id != user_account.username:
        # FIXME: respond nicely
        raise Exception(
            f"credentials do not match URL ('{acct_id}' vs '{user_account.username}')"
        )

    reserve_pub = expect_json_body_str(request, "reserve_pub")
    debit_account_payto = expect_json_body_str(request, "debit_account")
    amount = expect_json_body_amount(request, "amount")

    debit_account_name = get_acct_from_payto(debit_account_payto)
    LOGGER.info(
        f"adding incoming balance to exchange ({acct_id}) from account {debit_account_payto} ({debit_account_name})"
    )
    debit_user = User.objects.get(username=debit_account_name)
    debit_account = BankAccount.objects.get(user=debit_user)
    subject = f"{reserve_pub}"

    # check if currency is acceptable
    if amount.currency != settings.TALER_CURRENCY:
        return JsonResponse(
            {
                "code": 30, # TALER_EC_BANK_DUPLICATE_RESERVE_PUB_SUBJECT
                "hint": "The specified currency is not supported."
            },
            status=HTTPStatus.CONFLICT
        )

    # check if reserve pub exists already.
    try:
        BankTransaction.objects.get(subject=subject)

    except BankTransaction.DoesNotExist:

        wtrans = wire_transfer(
            amount,
            debit_account,
            exchange_account,
            subject)

        return JsonResponse(
            {
                "row_id": wtrans.id,
                "timestamp": dict(t_ms=(int(wtrans.date.timestamp()) * 1000)),
            }
        )
    # Here means this public key was used already: must fail.
    return JsonResponse(
        {
            "code": 5114, # TALER_EC_BANK_DUPLICATE_RESERVE_PUB_SUBJECT
            "hint": "The reserve public cannot be used multiple times."
        },
        status=HTTPStatus.CONFLICT
    )

@csrf_exempt
@require_POST
@login_via_headers
def twg_transfer(request, user_account, acct_id):
    """
    Transfer money from the exchange to a merchant account.
    """

    exchange_account = user_account.bankaccount

    if acct_id != user_account.username:
        # FIXME: respond nicely
        raise Exception(
            f"credentials do not match URL ('{acct_id}' vs '{user_account.username}')"
        )

    request_uid = expect_json_body_str(request, "request_uid")
    wtid = expect_json_body_str(request, "wtid")
    amount = expect_json_body_amount(request, "amount")
    exchange_base_url = expect_json_body_str(request, "exchange_base_url")
    credit_account_payto = expect_json_body_str(request, "credit_account")

    credit_account_name = get_acct_from_payto(credit_account_payto)
    try:
        credit_user = User.objects.get(username=credit_account_name)
    except User.DoesNotExist:
        LOGGER.error(f"credit account '{credit_account_name}' does not exist")
        # FIXME: use EC from taler-util library
        return JsonResponse(
            dict(
                code=ErrorCode.BANK_UNKNOWN_ACCOUNT,
                error="credit account does not exist",
            ),
            status=404,
        )
    credit_account = BankAccount.objects.get(user=credit_user)

    subject = f"{wtid} {exchange_base_url}"

    wtrans = wire_transfer(
        amount, exchange_account, credit_account, subject, request_uid
    )

    return JsonResponse(
        {
            "row_id": wtrans.id,
            "timestamp": dict(t_ms=(int(wtrans.date.timestamp()) * 1000)),
        }
    )


def get_plain_host(request):
    h = request.META.get("HTTP_HOST", "localhost")
    # remove port
    return h.split(":")[0]


def get_payto_from_account(request, acct):
    h = get_plain_host(request)
    return f"payto://x-taler-bank/{h}/{acct.user.username}"


def get_reserve_pub(subject):
    # obey to regex: \\b[a-z0-9A-Z]{52}\\b
    regex = re.compile(r"\b[a-z0-9A-Z]{52}\b")
    ret = regex.search(subject)
    if ret:
        return ret.group(0)
    return None


@require_GET
@login_via_headers
def twg_history_incoming(request, user_account, acct_id):
    history = []
    delta = int(request.GET["delta"])
    start_str = request.GET.get("start")
    if start_str is None:
        start = None
    else:
        start = int(start_str)
    qs = query_history(
        user_account.bankaccount,
        "credit",
        delta,
        start,
    )
    for item in qs:
        rp = get_reserve_pub(item.subject)
        if rp is None:
            continue
        history.append(
            dict(
                row_id=item.id,
                amount=item.amount.stringify(settings.TALER_DIGITS),
                date=dict(t_ms=(int(item.date.timestamp()) * 1000)),
                reserve_pub=rp,
                credit_account=get_payto_from_account(request, item.credit_account),
                debit_account=get_payto_from_account(request, item.debit_account),
            )
        )
    return JsonResponse(dict(incoming_transactions=history), status=HTTPStatus.OK)


@require_GET
@login_via_headers
def twg_history_outgoing(request, user_account, acct_id):
    history = []
    delta = int(request.GET["delta"])
    start_str = request.GET.get("start")
    if start_str is None:
        start = None
    else:
        start = int(start_str)
    qs = query_history(
        user_account.bankaccount,
        "debit",
        delta,
        start,
    )
    for item in qs:
        # FIXME: proper parsing, more structure in subject
        wtid, exchange_base_url = item.subject.split(" ")
        history.append(
            dict(
                row_id=item.id,
                amount=item.amount.stringify(settings.TALER_DIGITS),
                date=dict(t_ms=(int(item.date.timestamp()) * 1000)),
                wtid=wtid,
                exchange_base_url=exchange_base_url,
                credit_account=get_payto_from_account(request, item.credit_account),
                debit_account=get_payto_from_account(request, item.debit_account),
            )
        )
    return JsonResponse(dict(outgoing_transactions=history), status=HTTPStatus.OK)


##
# Implements the HTTP basic auth schema.
#
# @param request Django-specific HTTP request object.
# @return Django-specific "authentication object".
def basic_auth(request):
    auth_header = request.META.get("HTTP_AUTHORIZATION")

    if not auth_header:
        raise LoginFailed("missing Authorization header")

    tokens = auth_header.split(" ")
    if len(tokens) != 2:
        raise LoginFailed("invalid Authorization header")

    # decode the base64 content.
    if tokens[0] != "Basic":
        raise LoginFailed("Not supporting '%s' authorization method" % tokens[0])

    username, password = base64.b64decode(tokens[1]).decode("utf-8").split(":")
    return django.contrib.auth.authenticate(username=username, password=password)


def make_taler_withdraw_uri(request, withdraw_id):
    if request.is_secure():
        proto_extra = ""
    else:
        proto_extra = "+http"
    pfx = get_script_prefix().strip("/")
    if len(pfx) == 0:
        pfx_components = []
    else:
        pfx_components = pfx.split("/")
    host = request.get_host()
    p = "/".join([host] + pfx_components + ["api"] + [str(withdraw_id)])
    return f"taler{proto_extra}://withdraw/{p}"


@login_via_headers
@csrf_exempt
@require_POST
def withdraw_headless(request, user):
    """
    Serves a headless withdrawal request for the Taler protocol.
    """
    data = WithdrawHeadless(json.loads(decode_body(request)))
    h = get_plain_host(request)
    sender_payto = f"payto://x-taler-bank/{h}/{user.username}"
    ret_obj = {"sender_wire_details": sender_payto}

    exchange_payto = data.get("exchange_payto_uri")
    if not exchange_payto:
        return JsonResponse(
            dict(hint="exchange_payto_uri missig"), status=HTTPStatus.BAD_REQUEST
        )
    exchange_account_name = get_acct_from_payto(exchange_payto)
    try:
        exchange_user = User.objects.get(username=exchange_account_name)
    except User.DoesNotExist:
        return JsonResponse(
            dict(hint="exchange bank account does not exist"),
            status=HTTPStatus.NOT_FOUND,
        )
    exchange_bankaccount = exchange_user.bankaccount
    wire_transfer(
        Amount.parse(data.get("amount")),
        user.bankaccount,
        exchange_bankaccount,
        data.get("reserve_pub"),
    )

    return JsonResponse(ret_obj)


@csrf_exempt
@allow_origin_star
def api_withdraw_operation(request, withdraw_id):
    """
    Endpoint used by the browser and wallet to check withdraw status and
    put in the exchange info.
    """
    try:
        op = TalerWithdrawOperation.objects.get(withdraw_id=withdraw_id)
    except ObjectDoesNotExist:
        return JsonResponse(
            dict(error="withdraw operation does not exist"), status=HTTPStatus.NOT_FOUND
        )

    if request.method == "POST":
        data = json.loads(decode_body(request))
        exchange_payto_uri = data.get("selected_exchange")
        try:
            exchange_account_name = get_acct_from_payto(exchange_payto_uri)
        except:
            return JsonResponse(
                dict(error="exchange payto URI malformed"),
                status=HTTPStatus.BAD_REQUEST,
            )
        try:
            exchange_user = User.objects.get(username=exchange_account_name)
        except User.DoesNotExist:
            return JsonResponse(
                dict(
                    code=ErrorCode.BANK_UNKNOWN_ACCOUNT,
                    hint="bank account in payto URI unknown",
                ),
                status=HTTPStatus.BAD_REQUEST,  # WTF? NOT_FOUND?
            )
        exchange_account = exchange_user.bankaccount
        selected_reserve_pub = data.get("reserve_pub")
        if not isinstance(selected_reserve_pub, str):
            return JsonResponse(
                dict(error="reserve_pub must be a string"),
                status=HTTPStatus.BAD_REQUEST,
            )
        if op.selection_done:
            if (
                op.selected_exchange_account != exchange_account
                or op.selected_reserve_pub != selected_reserve_pub
            ):
                return JsonResponse(
                    dict(
                        code=ErrorCode.BANK_WITHDRAWAL_OPERATION_RESERVE_SELECTION_CONFLICT,
                        hint="selection of withdraw parameters already done",
                    ),
                    status=HTTPStatus.CONFLICT,
                )
        else:
            with transaction.atomic():
                op.selected_exchange_account = exchange_account
                op.selected_reserve_pub = selected_reserve_pub
                if op.confirmation_done and not op.selection_done:
                    # Confirmation already happened, we still need to transfer funds!
                    wire_transfer(
                        op.amount,
                        op.withdraw_account,
                        op.selected_exchange_account,
                        op.selected_reserve_pub,
                    )
                op.selection_done = True
                op.save()
        return JsonResponse(
            dict(
                transfer_done=op.confirmation_done,
                confirm_transfer_url=request.build_absolute_uri(
                    reverse("withdraw-confirm", args=(withdraw_id,))
                ),
            )
        )
    elif request.method == "GET":
        host = request.get_host()
        return JsonResponse(
            dict(
                selection_done=op.selection_done,
                transfer_done=op.confirmation_done,
                aborted=op.aborted,
                amount=op.amount.stringify(),
                wire_types=["x-taler-bank"],
                sender_wire=f"payto://x-taler-bank/{host}/{op.withdraw_account.user.username}",
                suggested_exchange=settings.TALER_SUGGESTED_EXCHANGE,
                confirm_transfer_url=request.build_absolute_uri(
                    reverse("withdraw-confirm", args=(withdraw_id,))
                ),
            )
        )
    else:
        return JsonResponse(
            dict(error="only GET and POST are allowed"),
            status=HTTPStatus.METHOD_NOT_ALLOWED,
        )


@login_required
@require_POST
def start_withdrawal(request):
    """
    Serve a Taler withdrawal request; takes the amount chosen
    by the user, and builds a response to trigger the wallet into
    the withdrawal protocol
    """
    user_account = BankAccount.objects.get(user=request.user)
    amount_num_str = request.POST.get("withdraw-amount")
    amount = Amount.parse(f"{settings.TALER_CURRENCY}:{amount_num_str}")
    withdraw_amount = SignedAmount(True, amount)
    debt_threshold = SignedAmount.parse(settings.TALER_MAX_DEBT)
    user_balance = user_account.balance
    if user_balance - withdraw_amount < -debt_threshold:
        raise DebitLimitException(
            f"Aborting payment initiated by '{user_account.user.username}', debit limit {debt_threshold} crossed."
        )
    op = TalerWithdrawOperation(amount=amount, withdraw_account=user_account)
    op.save()
    return redirect("withdraw-show", withdraw_id=op.withdraw_id)


def get_qrcode_svg(data):
    factory = qrcode.image.svg.SvgImage
    img = qrcode.make(data, image_factory=factory)
    el = img.get_image()
    el.attrib["style"] = "shape-rendering: crispedges;"
    return lxml.etree.tostring(el).decode("utf-8")


@login_required
@require_GET
def show_withdrawal(request, withdraw_id):
    op = TalerWithdrawOperation.objects.get(withdraw_id=withdraw_id)
    if op.selection_done:
        return redirect("withdraw-confirm", withdraw_id=op.withdraw_id)
    taler_withdraw_uri = make_taler_withdraw_uri(request, op.withdraw_id)
    qrcode_svg = get_qrcode_svg(taler_withdraw_uri)
    context = dict(
        taler_withdraw_uri=taler_withdraw_uri,
        qrcode_svg=qrcode_svg,
        withdraw_check_url=reverse(
            "api-withdraw-operation", kwargs=dict(withdraw_id=op.withdraw_id)
        ),
    )
    resp = render(request, "withdraw_show.html", context, status=402)
    resp["Taler"] = taler_withdraw_uri
    return resp


@login_required
@require_http_methods(["GET", "POST"])
def confirm_withdrawal(request, withdraw_id):
    op = TalerWithdrawOperation.objects.get(withdraw_id=withdraw_id)
    if not op.selection_done:
        raise Exception("invalid state (withdrawal parameter selection not done)")
    if op.confirmation_done:
        return redirect("profile")
    if request.method == "POST":
        hashed_attempt = hash_answer(request.POST.get("pin_0", ""))
        hashed_solution = request.POST.get("pin_1", "")
        if hashed_attempt != hashed_solution:
            LOGGER.warning(
                "Wrong CAPTCHA answer: %s vs %s",
                type(hashed_attempt),
                type(request.POST.get("pin_1")),
            )
            set_session_hint(
                request, success=False, hint=gettext("Wrong CAPTCHA answer.")
            )
            return redirect("withdraw-confirm", withdraw_id=withdraw_id)
        op.confirmation_done = True
        op.save()
        wire_transfer(
            op.amount,
            BankAccount.objects.get(user=request.user),
            op.selected_exchange_account,
            op.selected_reserve_pub,
        )
        set_session_hint(request, success=True, hint=gettext("Withdrawal successful!"))
        request.session["just_withdrawn"] = True
        return redirect("profile")

    if request.method == "GET":
        question, hashed_answer = make_question()
        is_success, hint = get_session_hint(request)
        context = dict(
            question=question,
            hashed_answer=hashed_answer,
            withdraw_id=withdraw_id,
            account_id=request.user.get_username(),
            amount=op.amount.stringify(settings.TALER_DIGITS),
            exchange=op.selected_exchange_account.user,
            is_success=is_success,
            hint=hint,
        )
        return render(request, "withdraw_confirm.html", context)
    raise Exception("not reached")


def wire_transfer(amount, debit_account, credit_account, subject, request_uid=None):
    """
    Make a wire transfer between two accounts of this demo bank.
    """
    if debit_account.pk == credit_account.pk:
        LOGGER.error("Debit and credit account are the same!")
        raise SameAccountException()

    if request_uid is None:
        request_uid = str(uuid.uuid4())
        pass
    else:
        # check for existing transfer
        try:
            etx = BankTransaction.objects.get(request_uid=request_uid)
        except BankTransaction.DoesNotExist:
            # We're good, no existing transaction with the same request_uid exists
            pass
        else:
            if (
                etx.amount != amount
                or etx.debit_account != debit_account
                or etx.credit_account != debit_account
                or etx.subject != subject
            ):
                return JsonResponse(
                    data=dict(
                        hint="conflicting transfer with same request_uid exists",
                        ec=ErrorCode.BANK_WITHDRAWAL_OPERATION_RESERVE_SELECTION_CONFLICT,
                    ),
                    status=HTTPStatus.CONFLICT,
                )

    LOGGER.info(
        "transferring %s => %s, %s, %s"
        % (
            debit_account.user.username,
            credit_account.user.username,
            amount.stringify(),
            subject,
        )
    )

    transaction_item = BankTransaction(
        amount=amount,
        credit_account=credit_account,
        debit_account=debit_account,
        subject=subject,
        request_uid=request_uid,
    )

    if debit_account.user.username == "Bank":
        threshold = -SignedAmount.parse(settings.TALER_MAX_DEBT_BANK)
    else:
        threshold = -SignedAmount.parse(settings.TALER_MAX_DEBT)

    if debit_account.balance - SignedAmount(True, amount) < threshold:
        raise DebitLimitException(
            f"Aborting payment initiated by '{debit_account.user.username}', debit limit {threshold} crossed."
        )

    debit_account.balance -= SignedAmount(True, amount)
    credit_account.balance += SignedAmount(True, amount)

    with transaction.atomic():
        debit_account.save()
        credit_account.save()
        transaction_item.save()

    return transaction_item


@csrf_exempt
@require_GET
@login_via_headers
def bank_accounts_api_balance(request, user_account, acct_id):
    """
    Query the balance for an account.
    """
    acct = user_account.bankaccount

    if acct_id != user_account.username:
        # FIXME: respond nicely
        raise Exception(
            f"credentials do not match URL ('{acct_id}' vs '{user_account.username}')"
        )

    return JsonResponse(
        dict(
            balance=dict(
                amount=acct.balance.amount.stringify(),
                credit_debit_indicator=(
                    "credit" if acct.balance.is_positive else "debit"
                ),
            )
        )
    )


@csrf_exempt
@require_POST
@login_via_headers
def bank_accounts_api_create_withdrawal(request, user, acct_id):
    user_account = BankAccount.objects.get(user=user)

    if acct_id != user_account.user.username:
        # FIXME: respond nicely
        raise Exception(
            f"credentials do not match URL ('{acct_id}' vs '{user_account.username}')"
        )

    data = WithdrawHeadlessUri(json.loads(decode_body(request)))
    amount = Amount.parse(data.get("amount"))
    withdraw_amount = SignedAmount(True, amount)
    debt_threshold = SignedAmount.parse(settings.TALER_MAX_DEBT)
    user_balance = user_account.balance
    if user_balance - withdraw_amount < -debt_threshold:
        raise DebitLimitException(
            f"Aborting payment initiated by '{user_account.user.username}', debit limit {debt_threshold} crossed."
        )
    op = TalerWithdrawOperation(amount=amount, withdraw_account=user_account)
    op.save()
    taler_withdraw_uri = make_taler_withdraw_uri(request, op.withdraw_id)
    return JsonResponse(
        {"taler_withdraw_uri": taler_withdraw_uri, "withdrawal_id": op.withdraw_id}
    )


@csrf_exempt
@require_GET
@login_via_headers
def bank_accounts_api_get_withdrawal(request, user, acct_id, wid):
    user_account = BankAccount.objects.get(user=user)
    if acct_id != user_account.user.username:
        # FIXME: respond nicely
        raise Exception(
            f"credentials do not match URL ('{acct_id}' vs '{user_account.username}')"
        )
    op = TalerWithdrawOperation.objects.get(withdraw_id=wid)
    selected_exchange_account = None
    if op.selected_exchange_account:
        selected_exchange_account = op.selected_exchange_account.user.username
    return JsonResponse(
        {
            "amount": op.amount.stringify(),
            "selection_done": op.selection_done,
            "confirmation_done": op.confirmation_done,
            "selected_reserve_pub": op.selected_reserve_pub,
            "selected_exchange_account": selected_exchange_account,
            "aborted": op.aborted,
        }
    )


def withdraw_abort_internal(wid):
    op = TalerWithdrawOperation.objects.get(withdraw_id=wid)
    if op.confirmation_done:
        return dict(status=HTTPStauts.CONFLICT, hint="can't abort confirmed withdrawal")
    op.aborted = True
    op.save()
    return dict(status=HTTPStatus.OK, hint="withdraw successfully aborted")


@require_POST
@login_required
def abort_withdrawal(request, withdraw_id):
    internal_status = withdraw_abort_internal(withdraw_id)
    set_session_hint(
        request,
        success=internal_status["status"] == HTTPStatus.OK,
        hint=internal_status["hint"],
    )
    return redirect("profile")


@csrf_exempt
@require_POST
@login_via_headers
def bank_accounts_api_abort_withdrawal(request, user, acct_id, wid):
    user_account = BankAccount.objects.get(user=user)
    if acct_id != user_account.user.username:
        # FIXME: respond nicely
        raise Exception(
            f"credentials do not match URL ('{acct_id}' vs '{user_account.username}')"
        )
    internal_status = withdraw_abort_internal(wid)
    return JsonResponse(
        dict(hint=internal_status["hint"]), status=internal_status["status"]
    )


@csrf_exempt
@require_POST
@login_via_headers
def bank_accounts_api_confirm_withdrawal(request, user, acct_id, wid):
    user_account = BankAccount.objects.get(user=user)
    if acct_id != user_account.user.username:
        # FIXME: respond nicely
        raise Exception(
            f"credentials do not match URL ('{acct_id}' vs '{user_account.username}')"
        )
    op = TalerWithdrawOperation.objects.get(withdraw_id=wid)
    if op.confirmation_done:
        return JsonResponse(dict(), status=HTTPStatus.OK)
    if op.aborted:
        return JsonResponse(
            dict(hint="can't confirm aborted withdrawal"), status=HTTPStatus.CONFLICT
        )

    with transaction.atomic():
        if op.selection_done:
            wire_transfer(
                op.amount,
                user_account,
                op.selected_exchange_account,
                op.selected_reserve_pub,
            )
        op.confirmation_done = True
        op.save()
    return JsonResponse(dict(), status=HTTPStatus.OK)



##
# This file is part of TALER
# (C) 2017 Taler Systems SA
#
# TALER is free software; you can redistribute it and/or
# modify it under the terms of the GNU Affero General Public
# License as published by the Free Software Foundation; either
# version 3, or (at your option) any later version.
#
# TALER is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with TALER; see the file COPYING.  If not,
# see <http://www.gnu.org/licenses/>
#
#  @author Florian Dold
#  @file CLI tool to manage all the bank's tasks.

import argparse
import django
import sys
import os
import os.path
import site
import logging
import inspect
import click

from taler.util.talerconfig import TalerConfig
from django.core.management import call_command

SITE_PACKAGES = os.path.abspath(os.path.dirname(__file__) + "/..")

LOGGER = logging.getLogger(__name__)

# No perfect match to our logging format, but good enough ...
UWSGI_LOGFMT = "%(ltime) %(proto) %(method) %(uri) %(proto) => %(status)"
os.environ.setdefault("DJANGO_SETTINGS_MODULE", "talerbank.settings")

# Argument to tell uWSGI to load the python plugin.
# This hack is required, because on systems where the plugin is statically linked,
# loading it causes an error.
arg_load_python = "--if-not-plugin python --plugins python --endif".split(" ")


@click.group(help="Manager script of Taler bank.")
@click.pass_context
@click.option(
    "--http-port",
    help="Set HTTP as the serving protocol (taking precedence over the config.), and set the port.",
    type=int,
)
@click.option("-c", "--config", help="Path to the config file.")
@click.option("--with-db", help="Database connection string.")
def cli(ctx, http_port, config, with_db):
    if with_db:
        os.environ.setdefault("TALER_BANK_ALTDB", with_db)
    if config:
        os.environ["TALER_CONFIG_FILE"] = config

    ctx.obj = dict(http_port=http_port, config=config)


@cli.command(help="Serve the bank")
@click.pass_obj
def serve(obj):
    # if --http-port option is found, then serve via HTTP.
    # Otherwise serve on whatever protocol is specified in the config.
    serve = "http"
    if not obj["http_port"]:
        TC = TalerConfig.from_file(os.environ.get("TALER_CONFIG_FILE"))
        serve = TC["bank"]["serve"].value_string(required=True).lower()

    if serve == "http":
        return handle_serve_http(obj["http_port"])
    handle_serve_uwsgi()


@cli.command(
    name="django",
    help="Invoke 'django' sub-commands",
    context_settings=dict(ignore_unknown_options=True),
)
@click.argument("args", nargs=-1, type=click.UNPROCESSED)
def handle_django(args):
    django.setup()
    # always run 'migrate' first, in case a virgin db is being used.
    call_command("migrate")
    django_index = sys.argv.index("django")
    from django.core.management import execute_from_command_line

    execute_from_command_line([sys.argv[0] + " django"] + sys.argv[django_index + 1 :])


def handle_serve_http(port):
    import django

    django.setup()
    print("migrating")
    call_command("migrate")
    print("providing accounts")
    call_command("provide_accounts")
    print("checking")
    call_command("check")
    if port is None:
        TC = TalerConfig.from_file(os.environ.get("TALER_CONFIG_FILE"))
        port = TC["bank"]["http_port"].value_int(required=True)

    httpspec = ":%d" % (port,)
    params = [
        "uwsgi",
        "uwsgi",
        *arg_load_python,
        "--static-map",
        "/static=%s/talerbank/app/static" % SITE_PACKAGES,
        "--die-on-term",
        "--no-orphans",
        "--master",
        "--http",
        httpspec,
        "--log-format",
        UWSGI_LOGFMT,
        "--module",
        "talerbank.wsgi",
    ]
    os.execlp(*params)


def handle_serve_uwsgi():
    django.setup()
    call_command("migrate")
    call_command("provide_accounts")
    call_command("check")
    TC = TalerConfig.from_file(os.environ.get("TALER_CONFIG_FILE"))
    serve_uwsgi = TC["bank"]["uwsgi_serve"].value_string(required=True).lower()
    params = [
        "uwsgi",
        "uwsgi",
        *arg_load_python,
        "--static-map",
        "/static=%s/talerbank/app/static" % SITE_PACKAGES,
        "--die-on-term",
        "--no-orphans",
        "--master",
        "--log-format",
        UWSGI_LOGFMT,
        "--module",
        "talerbank.wsgi",
    ]
    if serve_uwsgi == "tcp":
        port = TC["bank"]["uwsgi_port"].value_int(required=True)
        spec = ":%d" % (port,)
        params.extend(["--socket", spec])
    else:
        spec = TC["bank"]["uwsgi_unixpath"].value_filename(required=True)
        mode = TC["bank"]["uwsgi_unixpath_mode"].value_filename(required=True)
        params.extend(["--socket", spec])
        params.extend(["--chmod-socket=" + mode])
        try:
            os.makedirs(os.path.dirname(spec), exist_ok=True)
        except FileNotFoundError:
            print(f"{spec} is not a valid file path.")
            sys.exit(1)
    logging.info("launching uwsgi with argv %s", params[1:])
    os.execlp(*params)


@cli.command(help="Print config values.")
def config():
    TC = TalerConfig.from_file(os.environ.get("TALER_CONFIG_FILE"))
    TC.dump()

def run():
    cli()

if __name__ == "__main__":
    cli()



"""
Django settings for talerbank.

For more information on this file, see
https://docs.djangoproject.com/en/1.9/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/1.9/ref/settings/
"""

import base64
import logging
import os
import re
import sys
import urllib.parse
from taler.util.talerconfig import TalerConfig, ConfigurationError

LOGGER = logging.getLogger(__name__)

LOGGER.info(
    "DJANGO_SETTINGS_MODULE: %s" % os.environ.get("DJANGO_SETTINGS_MODULE")
)

TC = TalerConfig.from_file(os.environ.get("TALER_CONFIG_FILE"))

# Build paths inside the project like this:
# os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/1.9/howto/deployment/checklist/

SECRET_KEY = os.environ.get("TALER_BANK_SECRET_KEY", None)

if not SECRET_KEY:
    LOGGER.info(
        "secret key not configured in"
        " TALER_BANK_SECRET_KEY env variable,"
        " generating random secret"
    )
    SECRET_KEY = base64.b64encode(os.urandom(32)).decode("utf-8")

# SECURITY WARNING: don't run with debug turned on in production!

if "demo" == os.environ.get("TALER_ENV_NAME"):
    DEBUG = False
else:
    DEBUG = True

ADMIN_ENABLED = False

ALLOWED_HOSTS = ["*"]

LOGIN_URL = "login"

LOGIN_REDIRECT_URL = "index"

# Application definition

INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "talerbank.app",
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.locale.LocaleMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "talerbank.app.middleware.ExceptionMiddleware",
    "talerbank.app.middleware.DecompressionMiddleware",
]

TEMPLATES = [{
    "BACKEND":
    "django.template.backends.jinja2.Jinja2",
    "DIRS": [
        os.path.join(BASE_DIR, "talerbank/app/templates"),
    ],
    "OPTIONS": {
        "environment": "talerbank.jinja2.environment",
        "context_processors": ["talerbank.jinja2.context_processor"],
    },
}]

# Disable those, since they don't work with
# jinja2 anyways.
TEMPLATE_CONTEXT_PROCESSORS = []

WSGI_APPLICATION = "talerbank.wsgi.application"

# Database
# https://docs.djangoproject.com/en/1.9/ref/settings/#databases

DATABASES = {}

DBNAME = TC.value_string("bank", "database", required=True)
DBNAME = os.environ.get("TALER_BANK_ALTDB", DBNAME)

# Tells Django to add a BigAutoField column to every table that
# doesn't have a primary key explicitly defined.  Such column will
# then be the primary key.
DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"

if not DBNAME:
    raise Exception("DB not specified (neither in config or as" "cli argument)")

LOGGER.info("dbname: %s" % DBNAME)

CHECK_DBSTRING_FORMAT = re.search(
    r"[a-z]+:///[a-z]+([\?][a-z]+=[a-z/]+)?", DBNAME
)
if not CHECK_DBSTRING_FORMAT:
    LOGGER.error(
        "Bad db string given '%s', respect the format"
        "'dbtype:///dbname'" % DBNAME
    )
    sys.exit(2)

DBCONFIG = {}
# Maybe trust the parsing from urlparse?
DB_URL = urllib.parse.urlparse(DBNAME)

if DB_URL.scheme not in ("postgres") or DB_URL.scheme == "":
    LOGGER.error("DB '%s' is not supported" % DB_URL.scheme)
    sys.exit(1)
if DB_URL.scheme == "postgres":
    DBCONFIG["ENGINE"] = "django.db.backends.postgresql_psycopg2"
    DBCONFIG["NAME"] = DB_URL.path.lstrip("/")

if not DB_URL.netloc:
    P = urllib.parse.parse_qs(DB_URL.query)
    if ("host" not in P) or P["host"] == "":
        HOST = None
    else:
        HOST = P["host"][0]
else:
    HOST = DB_URL.netloc

if HOST:
    DBCONFIG["HOST"] = HOST  # Sockets directory.

DATABASES["default"] = DBCONFIG

# Password validation
# https://docs.djangoproject.com/en/1.9/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME":
        "django.contrib.auth.password_validation"
        ".UserAttributeSimilarityValidator"
    },
    {
        "NAME": "django.contrib.auth.password_validation"
        ".MinimumLengthValidator"
    },
    {
        "NAME": "django.contrib.auth.password_validation"
        ".CommonPasswordValidator"
    },
    {
        "NAME":
        "django.contrib.auth.password_validation"
        ".NumericPasswordValidator"
    },
]

# Internationalization
# https://docs.djangoproject.com/en/1.9/topics/i18n/

TIME_ZONE = "UTC"

USE_I18N = True

USE_L10N = True

USE_TZ = True

# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.9/howto/static-files/

STATIC_URL = "/static/"

STATICFILES_DIRS = [
    os.path.join(BASE_DIR, "talerbank/app/static"),
]

STATIC_ROOT = None
ROOT_URLCONF = "talerbank.urls"

try:
    TALER_CURRENCY = TC.value_string("taler", "currency", required=True)
except ConfigurationError as exc:
    LOGGER.error(exc)
    sys.exit(3)

TALER_MAX_DEBT = TC.value_string(
    "bank", "MAX_DEBT", default="%s:50.0" % TALER_CURRENCY
)
TALER_MAX_DEBT_BANK = TC.value_string(
    "bank", "MAX_DEBT_BANK", default="%s:0.0" % TALER_CURRENCY
)

TALER_DIGITS = TC.value_int("bank", "NDIGITS", default=2)
# Order matters
TALER_PREDEFINED_ACCOUNTS = [
    "Bank",
    "Exchange",
    "blog",
    "Tor",
    "GNUnet",
    "Taler",
    "FSF",
    "Tutorial",
    "Survey",
]
TALER_EXPECTS_DONATIONS = ["Tor", "GNUnet", "Taler", "FSF"]
TALER_SUGGESTED_EXCHANGE = TC.value_string("bank", "suggested_exchange")
TALER_SUGGESTED_EXCHANGE_PAYTO = TC.value_string(
    "bank", "suggested_exchange_payto"
)

_allow_reg = TC.value_string("bank", "ALLOW_REGISTRATIONS", default="no")

if _allow_reg.lower() == "yes":
    ALLOW_REGISTRATIONS = True
else:
    ALLOW_REGISTRATIONS = False

_show_freeform_withdrawal = TC.value_string(
    "bank", "SHOW_FREEFORM_WITHDRAWAL", default="no"
)
if _show_freeform_withdrawal.lower() == "yes":
    SHOW_FREEFORM_WITHDRAWAL = True
else:
    SHOW_FREEFORM_WITHDRAWAL = False



##
# This file is part of TALER
# (C) 2014, 2015, 2106 Taler Systems SA
#
# TALER is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 3,
# or (at your option) any later version.
#
# TALER is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with TALER; see the file COPYING.  If not,
# see <http://www.gnu.org/licenses/>
#
# @author Marcello Stanisci
# @author Florian Dold
# @brief Create the basic accounts to make the demo bank work.

import sys
import logging
from django.contrib.auth.models import User
from django.db.utils import ProgrammingError, OperationalError
from django.core.management.base import BaseCommand
from django.conf import settings
from ...models import BankAccount
from ...views import wire_transfer
from taler.util.amount import Amount
import uuid

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)


##
# Create a new bank account.
#
# @param username the username to associate to this account.
def make_account(username):
    try:
        User.objects.get(username=username)
    except User.DoesNotExist:
        LOGGER.info("Creating account for '%s'", username)
        BankAccount(
            user=User.objects.create_user(
                username=username, password=str(uuid.uuid4())
            ),
            is_public=True,
        ).save()

    except (OperationalError, ProgrammingError):
        LOGGER.error(
            "db does not exist, or the project"
            " is not migrated.  Try 'taler-bank-manage"
            " django migrate' in the latter case.",
            stack_info=False,
            exc_info=True,
        )
        sys.exit(1)


##
# Django-specific definition to register this command.
class Command(BaseCommand):
    help = "Provide initial user accounts"

    ##
    # Django-specific definition to invoke the account creator
    # @a make_account; it iterates over the list of basic accounts
    # (defined in the settings) and invoke the account creator
    # for each one of them.
    def handle(self, *args, **options):
        for username in settings.TALER_PREDEFINED_ACCOUNTS:
            make_account(username)



# This configure.py file is places in the public domain.

# Configure the build directory.
# This file is invoked by './configure' and should usually not be invoked
# manually.

import talerbuildconfig as tbc
import sys
import shutil

if getattr(tbc, "serialversion", 0) < 2:
    print("talerbuildconfig outdated, please update the build-common submodule and/or bootstrap")
    sys.exit(1)

b = tbc.BuildConfig()

# Declare dependencies
b.add_tool(tbc.PosixTool("find"))
b.add_tool(tbc.PosixTool("awk"))
b.add_tool(tbc.GenericTool("pip3", version_arg="--version"))
b.add_tool(tbc.GenericTool("poetry", version_arg="-V"))

b.run()

print("copying Makefile")
shutil.copyfile("build-system/Makefile", "Makefile")



##
# This file is part of TALER
# (C) 2014, 2015, 2016 Taler Systems SA
#
# TALER is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation; either version 3, or (at your
# option) any later version.
#
# TALER is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with TALER; see the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>
#
# @author Marcello Stanisci
# @brief CLI utility that issues a wire transfer.

import sys
import logging
import json
from django.core.management.base import BaseCommand
from django.contrib.auth import authenticate
from taler.util.amount import Amount
from ...views import wire_transfer, User
from ...models import BankAccount, BankTransaction

LOGGER = logging.getLogger(__name__)


##
# Django-specific definition to register the CLI utility.
class Command(BaseCommand):
    help = "Wire transfer money and return the transaction id."

    ##
    # Register the command line options of this command.
    #
    # @param self this object.
    # @param parser API used to actually register the option.
    def add_arguments(self, parser):
        parser.add_argument(
            "user",
            type=str,
            metavar="USERNAME",
            help="Which user is performing the wire transfer",
        )
        parser.add_argument(
            "password", type=str, metavar="PASSWORD", help="Performing user's password."
        )
        parser.add_argument(
            "credit-account",
            type=str,
            metavar="CREDIT-ACCOUNT",
            help="Which account will receive money.",
        )
        parser.add_argument(
            "subject",
            type=str,
            metavar="SUBJECT",
            help="SUBJECT will be the wire transfer subject.",
        )
        parser.add_argument(
            "amount",
            type=str,
            metavar="AMOUNT",
            help="Wire transfer's amount, given in the " "CURRENCY:X.Y form.",
        )

    ##
    # This callable gets invoked when the user invokes the
    # CLI utility; it is responsible of making the wire transfer
    # effective.
    #
    # @param self this object.
    # @param args arguments list -- currently unused.
    # @param options options given by the user at the command line.
    def handle(self, *args, **options):
        user = authenticate(username=options["user"], password=options["password"])
        if not user:
            LOGGER.error("Wrong user/password.")
            sys.exit(1)
        try:
            amount = Amount.parse(options["amount"])
        except BadFormatAmount:
            LOGGER.error("Amount's format is wrong: respect C:X.Y.")
            sys.exit(1)

        try:
            credit_account_user = User.objects.get(username=options["credit-account"])
            credit_account = credit_account_user.bankaccount
        except BankAccount.DoesNotExist:
            LOGGER.error("Credit account does not exist.")
            sys.exit(1)

        try:
            transaction = wire_transfer(
                amount, user.bankaccount, credit_account, options["subject"]
            )
            print("Transaction id: " + str(transaction.id))
        except Exception as exc:
            LOGGER.error(exc)
            sys.exit(1)



##
# This file is part of TALER
#  (C) 2014, 2015, 2016 Taler Systems SA
#
#  TALER is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation; either version 3, or
# (at your option) any later version. TALER is distributed in the
# hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License
# along with TALER; see the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>
#
#  @author Marcello Stanisci
#  @author Florian Dold

import uuid
from typing import Any, Tuple
from django.contrib.auth.models import User
from django.db import models
from django.conf import settings
from django.core.exceptions import ValidationError, ObjectDoesNotExist
from taler.util.amount import Amount, SignedAmount, CurrencyMismatchError


def get_zero_amount() -> Amount:
    """
    Helper function that instantiates a zero-valued Amount
    object in the currency that the bank runs on.
    """
    return Amount(settings.TALER_CURRENCY, 0, 0)


def get_zero_signed_amount() -> SignedAmount:
    """
    Helper function that instantiates a zero-valued SignedAmount
    object in the currency that the bank runs on.
    """
    return SignedAmount(True, get_zero_amount())


class SignedAmountField(models.Field):
    """Custom implementation of the SignedAmount class as a database type."""

    description = "Signed amount object in Taler style"

    def db_type(self, connection: Any) -> str:
        """
        Return the database type of the serialized amount.
        """
        return "varchar"

    def get_prep_value(self, value: SignedAmount) -> str:
        """
        Stringifies the Amount object to feed the DB connector.
        """
        c = value.amount.currency
        if settings.TALER_CURRENCY != c:
            raise CurrencyMismatchError(settings.TALER_CURRENCY, c)
        return value.stringify()

    @staticmethod
    def from_db_value(value: str, *args) -> Amount:
        """
        Parse the stringified Amount back to Python.

        Parameters
        ----------
        value : str
            Serialized amount coming from the database.
            (String in the usual CURRENCY:X.Y format)
        args : any
            Unused
        """
        del args  # pacify PEP checkers
        return SignedAmount.parse(value)

    def to_python(self, value: Any) -> Amount:
        """
        Parse the stringified Amount back to Python. FIXME:
        why this serializer consider _more_ cases respect to the
        one above ('from_db_value')?

        Parameters
        ----------
        value: serialized amount coming from the database

        """

        if isinstance(value, SignedAmount):
            return value
        try:
            return SignedAmount.parse(value)
        except BadFormatAmount:
            raise ValidationError(
                "Invalid input for a signed amount string: %s" % value
            )


class AmountField(models.Field):
    """Custom implementation of the Amount class as a database type."""

    description = "Amount object in Taler style"

    def db_type(self, connection: Any) -> str:
        """
        Return the database type of the serialized amount.
        """
        return "varchar"

    def get_prep_value(self, value: Amount) -> str:
        """
        Stringifies the Amount object to feed the DB connector.
        """
        if settings.TALER_CURRENCY != value.currency:
            raise CurrencyMismatchError(settings.TALER_CURRENCY, value.currency)
        return value.stringify()

    @staticmethod
    def from_db_value(value: str, *args) -> Amount:
        """
        Parse the stringified Amount back to Python.

        Parameters
        ----------
        value : str
            Serialized amount coming from the database.
            (String in the usual CURRENCY:X.Y format)
        args : any
            Unused
        """
        del args  # pacify PEP checkers
        return Amount.parse(value)

    def to_python(self, value: Any) -> Amount:
        """
        Parse the stringified Amount back to Python. FIXME:
        why this serializer consider _more_ cases respect to the
        one above ('from_db_value')?

        Parameters
        ----------
        value: serialized amount coming from the database

        """

        if isinstance(value, Amount):
            return value
        try:
            return Amount.parse(value)
        except BadFormatAmount:
            raise ValidationError("Invalid input for an amount string: %s" % value)


def join_dict(**inputDict):
    return ", ".join(["%s==%s" % (key, value) for (key, value) in inputDict.items()])


class BankAccount(models.Model):
    """
    The class representing a bank account.
    """

    is_public = models.BooleanField(default=False)
    account_no = models.AutoField(primary_key=True)
    user = models.OneToOneField(User, on_delete=models.CASCADE)
    balance = SignedAmountField(default=get_zero_signed_amount)


class BankTransaction(models.Model):
    """
    The class representing a bank transaction.
    """

    amount = AmountField(default=False)
    debit_account = models.ForeignKey(
        BankAccount,
        on_delete=models.CASCADE,
        db_index=True,
        related_name="debit_account",
    )
    credit_account = models.ForeignKey(
        BankAccount,
        on_delete=models.CASCADE,
        db_index=True,
        related_name="credit_account",
    )
    subject = models.CharField(default="(no subject given)", max_length=200)
    date = models.DateTimeField(auto_now=True, db_index=True)
    cancelled = models.BooleanField(default=False)
    request_uid = models.CharField(max_length=128, unique=True)


class TalerWithdrawOperation(models.Model):
    withdraw_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    amount = AmountField(default=False)
    withdraw_account = models.ForeignKey(
        BankAccount,
        on_delete=models.CASCADE,
        db_index=True,
        related_name="withdraw_account",
    )
    selection_done = models.BooleanField(default=False)
    confirmation_done = models.BooleanField(default=False)
    aborted = models.BooleanField(default=False)
    selected_exchange_account = models.ForeignKey(
        BankAccount,
        null=True,
        on_delete=models.CASCADE,
        related_name="selected_exchange_account",
    )
    selected_reserve_pub = models.TextField(null=True)



#  This file is part of TALER
#  (C) 2014, 2015, 2016 Taler Systems SA
#
#  TALER is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation; either version 3,
#  or (at your option) any later version.
#
#  TALER is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with TALER; see the file COPYING.  If not,
#  see <http://www.gnu.org/licenses/>.
#
#  @author Marcello Stanisci

import json
import time
import uuid
import zlib
import timeit
import logging
import unittest
import base64
from urllib.parse import unquote
from django.db import connection
from django.test import TestCase, Client
from django.urls import reverse
from django.conf import settings
from django.contrib.auth.models import User
from mock import patch, MagicMock
from .models import BankAccount, BankTransaction, TalerWithdrawOperation
from . import urls
from .views import wire_transfer, get_reserve_pub
from taler.util.payto import PaytoParse, PaytoFormatError
from taler.util.amount import (
    Amount,
    SignedAmount,
    CurrencyMismatchError,
    AmountFormatError,
)

LOGGER = logging.getLogger()
LOGGER.setLevel(logging.DEBUG)
LOGGER.handlers = []
# Logfile opens in 'append' mode.
fileHandler = logging.FileHandler("tests.log")
fileHandler.setLevel(logging.DEBUG)
fileHandler.setFormatter(
    logging.Formatter(fmt="%(asctime)-15s %(module)s %(levelname)s %(message)s")
)
LOGGER.addHandler(fileHandler)


def make_auth_line(username, password):
    credentials = "%s:%s" % (username, password)
    b64enc = base64.b64encode(bytes(credentials, "utf-8"))
    header_line = "Basic %s" % b64enc.decode()
    return header_line


def clear_db():
    User.objects.all().delete()
    BankAccount.objects.all().delete()
    BankTransaction.objects.all().delete()
    with connection.cursor() as cursor:
        cursor.execute("ALTER SEQUENCE app_bankaccount_account_no_seq" " RESTART")
        cursor.execute("ALTER SEQUENCE app_banktransaction_id_seq RESTART")


class WireGatewayTestCase(TestCase):
    def setUp(self):
        clear_db()
        exchange = User.objects.create_user(username="RandomExchange", password="XYZ")
        exchange.save()
        customer = User.objects.create_user(username="RandomCustomer", password="ABC")
        customer.save()
        exchange_bank_account = BankAccount(
            user=exchange,
            balance=SignedAmount(True, Amount(settings.TALER_CURRENCY, 100, 0)),
        )
        exchange_bank_account.save()
        customer_bank_account = BankAccount(
            user=customer,
            balance=SignedAmount(True, Amount(settings.TALER_CURRENCY, 100, 0)),
        )
        customer_bank_account.save()
        self.client = Client()

    def test_all(self):
        r = self.client.post(
            reverse("twg-add-incoming", kwargs=dict(acct_id="RandomExchange")),
            HTTP_AUTHORIZATION=make_auth_line("RandomExchange", "XYZ"),
            content_type="application/json",
            data=dict(
                amount=f"{settings.TALER_CURRENCY}:10",
                reserve_pub="FXWC2JHBY8B0XE2MMGAJ9TGPY307TN12HVEKYSTN6HE3GTHTF8XG",
                debit_account="payto://x-taler-bank/localhost/RandomCustomer",
            ),
        )
        self.assertEqual(r.status_code, 200)

        # Test incoming transfers of Exchange.
        r = self.client.get(
            reverse("twg-history-incoming", kwargs=dict(acct_id="RandomExchange")),
            dict(delta=5),
            HTTP_AUTHORIZATION=make_auth_line("RandomExchange", "XYZ"),
        )
        self.assertEqual(r.status_code, 200)

        # Test outgoing transfers of the Exchange.
        r = self.client.post(
            reverse("twg-transfer", kwargs=dict(acct_id="RandomExchange")),
            HTTP_AUTHORIZATION=make_auth_line("RandomExchange", "XYZ"),
            content_type="application/json",
            data=dict(
                request_uid="0",
                amount=f"{settings.TALER_CURRENCY}:3",
                exchange_base_url="mock",
                wtid="123",
                credit_account="payto://x-taler-bank/localhost/RandomCustomer",
            ),
        )
        r = self.client.get(
            reverse("twg-history-outgoing", kwargs=dict(acct_id="RandomExchange")),
            dict(delta=5),
            HTTP_AUTHORIZATION=make_auth_line("RandomExchange", "XYZ"),
        )
        self.assertEqual(r.status_code, 200)


class IntegrationApiTestCase(TestCase):
    def setUp(self):
        clear_db()

        self.exchange = User.objects.create_user(
            username="RandomExchange", password="XYZ"
        )
        self.exchange.save()
        self.exchange_bank_account = BankAccount(
            user=self.exchange,
            balance=SignedAmount(True, Amount(settings.TALER_CURRENCY, 100, 0)),
        )
        self.exchange_bank_account.save()

        self.user = User.objects.create_user(username="RandomUser", password="XYZ")
        self.user.save()
        self.user_bank_account = BankAccount(
            user=self.user,
            balance=SignedAmount(True, Amount(settings.TALER_CURRENCY, 100, 0)),
        )
        self.user_bank_account.save()
        self.client = Client()

    def test_config(self):
        c = Client()
        r = c.get("/config")
        self.assertEqual(r.status_code, 200)

    def test_withdraw(self):
        operation = TalerWithdrawOperation(
            amount=Amount(settings.TALER_CURRENCY, 100, 0),
            withdraw_account=self.user_bank_account,
        )
        operation.save()
        r = self.client.post(
            reverse(
                "api-withdraw-operation", kwargs=dict(withdraw_id=operation.withdraw_id)
            ),
            data=dict(
                reserve_pub="reserve-public-key",
                selected_exchange=f"payto://x-taler-bank/localhost/RandomExchange",
            ),
            content_type="application/json",
        )
        self.assertEqual(r.status_code, 200)


class AccessApiWithdrawTestCase(TestCase):
    def setUp(self):
        clear_db()
        self.user = User.objects.create_user(username="RandomUser", password="XYZ")
        self.user.save()
        self.user_bank_account = BankAccount(
            user=self.user,
            balance=SignedAmount(True, Amount(settings.TALER_CURRENCY, 100, 0)),
        )
        self.user_bank_account.save()
        self.client = Client()

    def create_withdrawal(self):
        r = self.client.post(
            reverse("access-api-withdrawal", kwargs=dict(acct_id="RandomUser")),
            HTTP_AUTHORIZATION=make_auth_line("RandomUser", "XYZ"),
            content_type="application/json",
            data=dict(amount=f"{settings.TALER_CURRENCY}:5"),
        )
        self.assertEqual(r.status_code, 200)
        data = r.content.decode("utf-8")
        data_dict = json.loads(data)
        withdrawal_id = data_dict.get("withdrawal_id")
        self.assertTrue(withdrawal_id)
        return withdrawal_id

    def test_accees_withdraw_create(self):
        self.create_withdrawal()

    def test_accees_withdraw_status(self):
        withdrawal_id = self.create_withdrawal()
        r = self.client.get(
            reverse(
                "access-api-withdrawal-status",
                kwargs=dict(acct_id="RandomUser", wid=withdrawal_id),
            ),
            HTTP_AUTHORIZATION=make_auth_line("RandomUser", "XYZ"),
        )
        self.assertEqual(r.status_code, 200)

    def test_accees_withdraw_abort(self):
        withdrawal_id = self.create_withdrawal()
        r = self.client.post(
            reverse(
                "access-api-withdrawal-abort",
                kwargs=dict(acct_id="RandomUser", wid=withdrawal_id),
            ),
            HTTP_AUTHORIZATION=make_auth_line("RandomUser", "XYZ"),
        )
        self.assertEqual(r.status_code, 200)

    def test_accees_withdraw_confirm(self):
        withdrawal_id = self.create_withdrawal()
        r = self.client.post(
            reverse(
                "access-api-withdrawal-confirm",
                kwargs=dict(acct_id="RandomUser", wid=withdrawal_id),
            ),
            HTTP_AUTHORIZATION=make_auth_line("RandomUser", "XYZ"),
        )
        self.assertEqual(r.status_code, 200)

    def test_accees_withdraw_abort_then_confirm(self):
        withdrawal_id = self.create_withdrawal()
        r = self.client.post(
            reverse(
                "access-api-withdrawal-abort",
                kwargs=dict(acct_id="RandomUser", wid=withdrawal_id),
            ),
            HTTP_AUTHORIZATION=make_auth_line("RandomUser", "XYZ"),
        )
        self.assertEqual(r.status_code, 200)
        r = self.client.post(
            reverse(
                "access-api-withdrawal-confirm",
                kwargs=dict(acct_id="RandomUser", wid=withdrawal_id),
            ),
            HTTP_AUTHORIZATION=make_auth_line("RandomUser", "XYZ"),
        )
        self.assertEqual(r.status_code, 409)

    def test_integration_api_withdraw_status(self):
        wid = self.create_withdrawal()
        r = self.client.get(
            reverse(
                "access-api-withdrawal-status",
                kwargs=dict(acct_id="RandomUser", wid=wid),
            ),
            HTTP_AUTHORIZATION=make_auth_line("RandomUser", "XYZ"),
        )
        self.assertEqual(r.status_code, 200)

    def test_integration_api_withdraw_confirm(self):
        wid = self.create_withdrawal()
        r = self.client.post(
            reverse(
                "access-api-withdrawal-confirm",
                kwargs=dict(acct_id="RandomUser", wid=wid),
            ),
            data=dict(
                reserve_pub="FXWC2JHBY8B0XE2MMGAJ9TGPY307TN12HVEKYSTN6HE3GTHTF8XG",
                selected_exchange="payto://x-taler-bank/localhost/RandomUser",
            ),
            content_type="application/json",
            HTTP_AUTHORIZATION=make_auth_line("RandomUser", "XYZ"),
        )
        self.assertEqual(r.status_code, 200)


class AccessApiBalanceTestCase(TestCase):
    def setUp(self):
        clear_db()
        self.user = User.objects.create_user(username="RandomUser", password="XYZ")
        self.user.save()
        self.user_bank_account = BankAccount(user=self.user)
        self.user_bank_account.save()

    def test_balance(self):
        c = Client()
        r = c.get(
            reverse("access-api-balance", kwargs=dict(acct_id="RandomUser")),
            HTTP_AUTHORIZATION=make_auth_line("RandomUser", "XYZ"),
        )
        self.assertEqual(r.status_code, 200)


class AccessApiTestingRegistrationTestCase(TestCase):
    def setUp(self):
        clear_db()
        self.user = User.objects.create_user(username="Bank", password="Bank")
        self.user.save()
        self.user_bank_account = BankAccount(user=self.user)
        self.user_bank_account.save()

    def test_testing_registration(self):
        c = Client()
        r = c.post(
            reverse("testing-withdraw-register"),
            data=dict(username="x", password="y"),
            content_type="application/json",
        )
        self.assertEqual(r.status_code, 200)


class ReservePubExtractionTestCase(TestCase):
    def test_extraction(self):
        self.assertTrue(
            get_reserve_pub("0T096A11M57GWGG0P6ZM9Z8G5829BFJFH2AN9R5T80FJ931DX7GG")
        )
        self.assertTrue(
            get_reserve_pub(
                "0T096A11M57GWGG0P6ZM9Z8G5829BFJFH2AN9R5T80FJ931DX7GG other data"
            )
        )
        self.assertFalse(get_reserve_pub("not a reserve public key"))


class PaytoParseTestCase(TestCase):
    def test_payto_wrong_protocol(self):
        self.assertRaises(PaytoFormatError, PaytoParse, "http://foo/bar")
    def test_payto_with_port_number(self):
        parsed = PaytoParse("payto://iban/localhost:1234/account")
        self.assertEqual(parsed.bank, "localhost:1234")
    def test_minimal(self):
        parsed = PaytoParse("payto://x-taler-bank/bank-hostname/Taler")
    def test_payto_malformed(self):
        self.assertRaises(PaytoFormatError, PaytoParse, "payto:foo/bar")
    def test_payto_noamount(self):
        parsed = PaytoParse(
            "payto://x-taler-bank/bank.int.taler.net/Exchange?message=0T096A11M57GWGG0P6ZM9Z8G5829BFJFH2AN9R5T80FJ931DX7GG"
        )
    def test_payto_parse(self):
        parsed = PaytoParse(
            "payto://x-taler-bank/bank.int.taler.net/Exchange?message=0T096A11M57GWGG0P6ZM9Z8G5829BFJFH2AN9R5T80FJ931DX7GG&amount=EUR:1"
        )
        self.assertEqual("Exchange", parsed.target)
        self.assertEqual("0T096A11M57GWGG0P6ZM9Z8G5829BFJFH2AN9R5T80FJ931DX7GG", parsed.message)
        self.assertEqual(parsed.amount.value, 1)
        self.assertEqual(parsed.amount.fraction, 0)
        self.assertEqual(parsed.amount.currency, "EUR")
        self.assertEqual(parsed.authority, "x-taler-bank")
        self.assertEqual(parsed.bank, "bank.int.taler.net")


class PublicAccountsTestCase(TestCase):
    def setUp(self):
        clear_db()
        self.user = User.objects.create_user(username="Bank", password="Bank")
        self.user.save()

        self.user_bank_account = BankAccount(
            account_no=100, is_public=True, user=self.user
        )

        self.user_bank_account.save()

    def test_public_accounts(self):
        self.assertTrue(User.objects.get(username="Bank"))

        response = self.client.get(reverse("public-accounts", urlconf=urls))


class WithdrawTestCase(TestCase):
    def setUp(self):
        self.user_bank_account = BankAccount(
            user=User.objects.create_user(
                username="test_user", password="test_password"
            ),
            account_no=100,
        )
        self.user_bank_account.save()

        self.exchange_bank_account = BankAccount(
            user=User.objects.create_user(username="test_exchange", password=""),
            account_no=99,
        )
        self.exchange_bank_account.save()
        self.client = Client()

    @patch("talerbank.app.views.wire_transfer")
    @patch("hashlib.new")
    @patch("time.time")
    @unittest.skip("skip outdated test case")
    def test_withdraw(self, mocked_time, mocked_hashlib, mocked_wire_transfer):
        amount = Amount(settings.TALER_CURRENCY, 0, 1)
        params = {
            "amount_value": str(amount.value),
            "amount_fraction": str(amount.fraction),
            "amount_currency": amount.currency,
            "reserve_pub": "UVZ789",
            "exchange": "https://exchange.example.com/",
            "exchange_wire_details": "payto://x-taler-bank/bank.example/99",
        }
        self.client.login(username="test_user", password="test_password")

        response = self.client.get(reverse("pin-question", urlconf=urls), params)
        self.assertEqual(response.status_code, 200)
        # We mock hashlib in order to fake the CAPTCHA.
        hasher = MagicMock()
        hasher.hexdigest = MagicMock()
        hasher.hexdigest.return_value = "0"
        mocked_hashlib.return_value = hasher
        mocked_time.return_value = 0

        response = self.client.post(
            reverse("withdraw-confirm", urlconf=urls), {"pin_1": "0"}
        )

        args, kwargs = mocked_wire_transfer.call_args
        del kwargs
        self.assertTrue(
            args[0].dump() == amount.dump()
            and self.user_bank_account in args
            and "UVZ789" in args
            and self.exchange_bank_account in args
        )

    def tearDown(self):
        clear_db()


class RegisterTestCase(TestCase):
    """User registration"""

    def setUp(self):
        clear_db()
        BankAccount(user=User.objects.create_user(username="Bank")).save()

    def tearDown(self):
        clear_db()

    def test_register(self):
        setattr(settings, "ALLOW_REGISTRATIONS", True)
        client = Client()
        response = client.post(
            reverse("register", urlconf=urls),
            {"username": "test_register", "password": "test_register"},
            follow=True,
        )
        self.assertIn(("/en/profile", 302), response.redirect_chain)
        # this assertion tests "/profile""s view
        self.assertEqual(200, response.status_code)

    def test_register_headless(self):
        setattr(settings, "ALLOW_REGISTRATIONS", True)
        client = Client()

        # Normal case.
        response = client.post(
            reverse("testing-withdraw-register", urlconf=urls),
            content_type="application/json",
            data={"username": "test_register_headless", "password": "password*+#@"},
        )
        self.assertEqual(200, response.status_code)

        # Double-check account creation.
        self.assertTrue(
            self.client.login(
                username="test_register_headless", password="password*+#@"
            )
        )

        # Try registering unavailable username.
        response = client.post(
            reverse("testing-withdraw-register", urlconf=urls),
            content_type="application/json",
            data={"username": "test_register_headless", "password": "password"},
        )
        self.assertEqual(409, response.status_code)

        # NOTE: Django 2.2.2 allows ANY character!  Is this normal?
        response = client.post(
            reverse("testing-withdraw-register", urlconf=urls),
            content_type="application/json",
            data={"username": "'''+++;;;'''", "password": "password2"},
        )
        self.assertEqual(200, response.status_code)


class LoginTestCase(TestCase):
    """User login"""

    def setUp(self):
        BankAccount(
            user=User.objects.create_user(
                username="test_user", password="test_password"
            )
        ).save()
        self.client = Client()

    def tearDown(self):
        clear_db()

    def test_login(self):
        self.assertTrue(
            self.client.login(username="test_user", password="test_password")
        )
        self.assertFalse(
            self.client.login(username="test_user", password="test_passwordii")
        )

    def test_failing_login(self):
        response = self.client.get(
            reverse("history", urlconf=urls),
            {"auth": "basic"},
            HTTP_AUTHORIZATION=make_auth_line("Wrong", "Credentials"),
        )
        data = response.content.decode("utf-8")
        self.assertEqual(401, response.status_code)


class AddIncomingTestCase(TestCase):
    """Test money transfer's API"""

    def setUp(self):
        BankAccount(
            user=User.objects.create_user(
                username="bank_user", password="bank_password"
            )
        ).save()
        BankAccount(
            user=User.objects.create_user(
                username="user_user", password="user_password"
            )
        ).save()

    def tearDown(self):
        clear_db()

    def test_add_incoming(self):
        client = Client()
        request_body = dict(
            reserve_pub="TESTWTID",
            amount=f"{settings.TALER_CURRENCY}:1.0",
            debit_account="payto://x-taler-bank/bank_user",
        )
        response = client.post(
            reverse("twg-add-incoming", urlconf=urls, args=["user_user"]),
            data=json.dumps(request_body),
            content_type="application/json",
            follow=True,
            HTTP_AUTHORIZATION=make_auth_line("user_user", "user_password"),
        )
        self.assertEqual(200, response.status_code)


class CustomDoesNotExistTestCase(TestCase):
    def test_bankaccount_doesnotexist(self):
        with self.assertRaises(BankAccount.DoesNotExist):
            BankAccount.objects.get(account_no=1000)
        with self.assertRaises(BankTransaction.DoesNotExist):
            BankTransaction.objects.get(subject="1000")


class HistoryTestCase(TestCase):
    def setUp(self):
        clear_db()
        debit_account = BankAccount(
            user=User.objects.create_user(username="User", password="Password"),
            balance=SignedAmount(True, Amount(settings.TALER_CURRENCY, 100, 0)),
        )
        debit_account.save()
        credit_account = BankAccount(
            user=User.objects.create_user(username="User0", password="Password0")
        )
        credit_account.save()
        for subject in ("a", "b", "c", "d", "e", "f", "g", "h", "i"):
            wire_transfer(
                Amount(settings.TALER_CURRENCY, 1, 0),
                debit_account,
                credit_account,
                subject,
            )

    def tearDown(self):
        clear_db()

    def test_history(self):
        def histquery(**urlargs):
            response = self.client.get(
                reverse("history", urlconf=urls),
                urlargs,
                HTTP_AUTHORIZATION=make_auth_line("User", "Password"),
            )
            return response

        # test query #1
        r = histquery(delta="-4", direction="both")
        rd = json.loads(r.content)
        self.assertEqual(r.status_code, 200)

        # test query #2
        r = histquery(delta="+1", start="5", direction="both")
        self.assertEqual(r.status_code, 200)
        rd = json.loads(r.content)
        self.assertEqual(r.status_code, 200)
        self.assertEqual(rd["data"][0]["row_id"], 6)

        # test query #3
        r = histquery(delta="+1", start="2", direction="both")
        self.assertEqual(r.status_code, 200)
        rd = json.loads(r.content)
        self.assertEqual(rd["data"][0]["wt_subject"], "c")

        # test query #4
        r = histquery(delta="-1", start="2", direction="both")
        self.assertEqual(r.status_code, 200)
        rd = json.loads(r.content)
        self.assertEqual(rd["data"][0]["wt_subject"], "a")

        # test query #5
        r = histquery(delta="1", start="11", direction="both")
        self.assertEqual(r.status_code, 200)
        rd = json.loads(r.content)
        self.assertEqual(len(rd["data"]), 0)


class DBCustomColumnTestCase(TestCase):
    def setUp(self):
        BankAccount(user=User.objects.create_user(username="U")).save()

    def tearDown(self):
        clear_db()

    def test_exists(self):
        user_bankaccount = BankAccount.objects.get(user=User.objects.get(username="U"))
        self.assertTrue(isinstance(user_bankaccount.balance, SignedAmount))


## This tests whether a bank account goes debit and then goes >=0
## again
class DebitTestCase(TestCase):
    def setUp(self):
        BankAccount(user=User.objects.create_user(username="U")).save()
        BankAccount(user=User.objects.create_user(username="U0")).save()

    def tearDown(self):
        clear_db()

    def test_green(self):
        user_bankaccount = BankAccount.objects.get(user=User.objects.get(username="U"))
        self.assertTrue(user_bankaccount.balance.is_zero())

    def test_red(self):
        user_bankaccount = BankAccount.objects.get(user=User.objects.get(username="U"))
        user_bankaccount0 = BankAccount.objects.get(
            user=User.objects.get(username="U0")
        )

        wire_transfer(
            Amount(settings.TALER_CURRENCY, 10, 0),
            user_bankaccount0,
            user_bankaccount,
            "Go green",
        )

        wire_transfer(
            Amount(settings.TALER_CURRENCY, 11, 0),
            user_bankaccount,
            user_bankaccount0,
            "Go red",
        )

        amt_one = SignedAmount.parse(f"{settings.TALER_CURRENCY}:1")

        self.assertEqual(user_bankaccount.balance, -amt_one)
        self.assertEqual(user_bankaccount0.balance, amt_one)


class MeasureHistory(TestCase):
    def setUp(self):
        self.user_bankaccount0 = BankAccount(
            user=User.objects.create_user(username="U0"),
            balance=SignedAmount(True, Amount(settings.TALER_CURRENCY, 3000, 0)),
        )
        self.user_bankaccount0.save()

        user_bankaccount = BankAccount(user=User.objects.create_user(username="U"))
        user_bankaccount.save()

        self.ntransfers = 1000

        # Make sure logging level is WARNING, otherwise the loop
        # will overwhelm the console.
        for i in range(self.ntransfers):
            del i  # to pacify PEP checkers
            wire_transfer(
                Amount(settings.TALER_CURRENCY, 1, 0),
                self.user_bankaccount0,
                user_bankaccount,
                "bulk",
            )

    def tearDown(self):
        clear_db()

    def test_extract_history(self):

        # Measure the time extract_history() needs to retrieve
        # ~ntransfers records.
        timer = timeit.Timer(
            stmt="extract_history(self.user_bankaccount0, False)",
            setup="from talerbank.app.views import extract_history",
            globals=locals(),
        )
        total_time = timer.timeit(number=1)
        allowed_time_per_record = 0.003
        self.assertLess(total_time, self.ntransfers * allowed_time_per_record)


class BalanceTestCase(TestCase):
    def setUp(self):
        self.the_bank = BankAccount(
            user=User.objects.create_user(username="U0", password="U0PASS"),
            balance=SignedAmount(True, Amount(settings.TALER_CURRENCY, 3, 0)),
        )
        self.the_bank.save()

        user = BankAccount(
            user=User.objects.create_user(username="U"),
            balance=SignedAmount(True, Amount(settings.TALER_CURRENCY, 10, 0)),
        )
        user.save()

        # bank: 3, user: 10 (START).

        # bank: 2, user: 11
        wire_transfer(
            Amount(settings.TALER_CURRENCY, 1, 0), self.the_bank, user, "mock"
        )

        # bank: 4, user: 9
        wire_transfer(
            Amount(settings.TALER_CURRENCY, 2, 0), user, self.the_bank, "mock"
        )

        # bank: -1, user: 14
        wire_transfer(
            Amount(settings.TALER_CURRENCY, 5, 0), self.the_bank, user, "mock"
        )

        # bank: 7, user: 6 (END)
        wire_transfer(
            Amount(settings.TALER_CURRENCY, 8, 0), user, self.the_bank, "mock"
        )

        # bank: -3, user: 16 (END)
        wire_transfer(
            Amount(settings.TALER_CURRENCY, 10, 0), self.the_bank, user, "mock"
        )

        self.client = Client()

    def tearDown(self):
        clear_db()

    def test_balance(self):
        self.client.login(username="U0", password="U0PASS")
        response = self.client.get(
            reverse("history", urlconf=urls),
            {"delta": -30, "direction": "both", "account_number": 55},
            HTTP_AUTHORIZATION=make_auth_line("U0", "U0PASS"),
        )
        data = response.content.decode("utf-8")
        self.assertEqual(response.status_code, 200)
        entries = json.loads(data)

        acc_bal = SignedAmount(True, Amount(settings.TALER_CURRENCY, 10, 0))

        for entry in reversed(entries["data"]):
            if entry["sign"] == "-":
                acc_bal += SignedAmount.parse(entry["amount"])
            if entry["sign"] == "+":
                acc_bal -= SignedAmount.parse(entry["amount"])

        expected_amount = SignedAmount.parse(f"{settings.TALER_CURRENCY}:16.0")
        self.assertEqual(acc_bal, expected_amount)



##
# This file is part of TALER
# (C) 2014, 2015, 2016 Taler Systems SA
#
# TALER is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation; either version 3, or (at your
# option) any later version.
#
# TALER is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with TALER; see the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>
#
# @author Marcello Stanisci
# @brief CLI utility to make wire transfers.

import sys
import logging
import json
from django.core.management.base import BaseCommand
from django.contrib.auth import authenticate
from taler.util.amount import Amount, AmountFormatError
from taler.util.payto import PaytoParse, PaytoFormatError
from ...views import wire_transfer, User
from ...models import BankAccount, BankTransaction

LOGGER = logging.getLogger(__name__)


##
# Django-specific definition to register the CLI utility.
class Command(BaseCommand):
    help = "Wire transfer money and return the transaction id."

    ##
    # Register the command line options of this command.
    #
    # @param self this object.
    # @param parser API used to actually register the option.
    def add_arguments(self, parser):
        parser.add_argument(
            "user",
            type=str,
            metavar="USERNAME",
            help="Which user is performing the wire transfer",
        )
        parser.add_argument(
            "password", type=str, metavar="PASSWORD", help="Performing user's password."
        )
        parser.add_argument(
            "payto-credit-account-with-subject",
            type=str,
            metavar="PAYTO-CREDIT-ACCOUNT-WITH-SUBJECT",
            help="Which account will receive money, in the form 'payto://x-taler-bank/[bank-host/]CreditAccountName?subject=PaymentSubject'.",
        )
        parser.add_argument(
            "amount",
            type=str,
            metavar="AMOUNT",
            help="Wire transfer's amount, given in the " "CURRENCY:X.Y form.",
        )

    ##
    # This callable gets invoked when the user invokes the
    # CLI utility; it is responsible of making the wire transfer
    # effective.
    #
    # @param self this object.
    # @param args arguments list -- currently unused.
    # @param options options given by the user at the command line.
    def handle(self, *args, **options):
        user = authenticate(username=options["user"], password=options["password"])
        if not user:
            LOGGER.error("Wrong user/password.")
            sys.exit(1)
        try:
            amount = Amount.parse(options["amount"])
        except AmountFormatError:
            print("Amount format is wrong: respect C:X.Y.")
            sys.exit(1)
        try:
            parsed_payto = PaytoParse(options["payto-credit-account-with-subject"])
            credit_account_user = User.objects.get(username=parsed_payto.target)
            credit_account = credit_account_user.bankaccount
        except User.DoesNotExist:
            print("Credit account does not exist.")
            sys.exit(1)
        except PaytoFormatError as e:
            print(e)
            sys.exit(1)
        if not parsed_payto.message:
            print("Please provide 'message' parameter along the payto URI")
            sys.exit(1)
        try:
            transaction = wire_transfer(
                amount, user.bankaccount, credit_account, parsed_payto.message
            )
            print("Transaction id: " + str(transaction.id))
        except Exception as exc:
            LOGGER.error(exc)
            sys.exit(1)



##
# This file is part of TALER
# (C) 2017 Taler Systems SA
#
# TALER is free software; you can redistribute it and/or
# modify it under the terms of the GNU Affero General Public
# License as published by the Free Software Foundation; either
# version 3, or (at your option) any later version.
#
# TALER is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with TALER; see the file COPYING.  If not,
# see <http://www.gnu.org/licenses/>
#
#  @author Florian Dold
#  @brief Extends the Jinja2 API with custom functions.

import os
import math
import json
from urllib.parse import urlparse
from django.urls import reverse, get_script_prefix
from django.conf import settings
from jinja2 import Environment
from django.utils.translation import gettext
import markupsafe


##
# Check if a URL is absolute or not.
#
# @param urloc the URL to check.
# @return True if the URL is absolute, False otherwise.
def is_absolute(urloc):
    return bool(urlparse(urloc).netloc)


##
# Join URL components held in a list, taking care
# of not having double slashes in the result.
#
# @param parts list of URL components.
# @return the string made of the joined components
def join_urlparts(*parts):
    ret = ""
    part = 0
    while part < len(parts):
        buf = parts[part]
        part += 1
        if ret.endswith("/"):
            buf = buf.lstrip("/")
        elif ret and not buf.startswith("/"):
            buf = "/" + buf
        ret += buf
    return ret


##
# Prefixes the argument with the location for static content.
#
# @param urloc the URL portion that should be prefixed; in
#        other words, this will be in final position in the
#        produced result.
# @return the URL that picks @a urloc from the location for
#         static content.
def static(urloc):
    if is_absolute(urloc):
        return urloc
    return join_urlparts(get_script_prefix(), settings.STATIC_URL, urloc)


##
# Helper function that fetches a value from the settings.
#
# @param name the name to lookup in the settings.
# @return @a name's value as defined in the settings, or
#         a empty string otherwise.
def settings_value(name):
    return getattr(settings, name, "")


##
# Fetch the URL given its "name".
#
# @param url_name URL's name as defined in urlargs.py
# @param kwargs key-value list that will be appended
#        to the URL as the parameter=value pairs.
def url(url_name, *args, **kwargs):
    # strangely, Django's 'reverse' function
    # takes a named parameter 'kwargs' instead
    # of real kwargs.
    return reverse(url_name, args=args, kwargs=kwargs)


##
# Helper function that reads a value from the environment.
#
# @param name env value to read
# @return the value, or None if not found.
def env_get(name, default=None):
    return os.environ.get(name, default)


##
# Jinja2 specific function used to activate the definitions
# of this module.
#
# @param options opaque argument (?) given from the caller.
# @return Jinja2-specific object that contains the new definitions.
def is_valid_amount(amount):
    if math.isnan(amount.value):
        return False
    return True


def tojson(x):
    """Convert object to json"""
    return json.dumps(x)


##
# Stringifies amount.
#
# @param amount amount object.
# @return amount pretty string.
def amount_stringify(amount):
    return amount.stringify(settings.TALER_DIGITS, pretty=True)


def get_locale(url):
    parts = url.split('/', 2)
    if (2 >= len(parts)):
        # Totally unexpected path format, do not localize
        return "en"
    lang = parts[1]
    return lang


all_languages = {
    "en": "English&nbsp;[en]",
    "ar": "Arabic&nbsp;[ar]",
    "zh_Hant": "Chinese&nbsp;[zh]",
    "fr": "French&nbsp;[fr]",
    "de": "German&nbsp;[de]",
    "hi": "Hindi&nbsp;[hi]",
    "it": "Italian&nbsp;[it]",
    "ja": "Japanese&nbsp;[ja]",
    "ko": "Korean&nbsp;[ko]",
    "pt": "Portuguese&nbsp;[pt]",
    "pt_BR": "Portuguese (Brazil)&nbsp;[pt_BR]",
    "ru": "Russian&nbsp;[ru]",
    "es": "Spanish&nbsp;[es]",
    "sv": "Swedish&nbsp;[sv]",
    "tr": "Turkish&nbsp;[tr]",
}


def context_processor(request):
    def getlang():
        return get_locale(request.path)

    return dict(getlang=getlang)


def environment(**options):
    env = Environment(**options)

    gettext_markup = lambda *args, **kwargs: markupsafe.Markup(
        gettext(*args, **kwargs)
    )

    env.globals.update({
        "static": static,
        "url": url,
        "settings_value": settings_value,
        "env": env_get,
        "is_valid_amount": is_valid_amount,
        "amount_stringify": amount_stringify,
        "tojson": tojson,
        "_": gettext_markup,
        "gettext": gettext_markup,
        "get_locale": get_locale,
        "getactive": lambda: "bank",
        "all_languages": all_languages,
    })
    return env



##
# This file is part of TALER
# (C) 2014, 2015, 2016, 2020 Taler Systems SA
#
#  TALER is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation; either version 3, or
# (at your option) any later version.
#
#  TALER is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public
# License along with TALER; see the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>
#
#  @author Marcello Stanisci
#  @brief definitions of JSON schemas for validating data

import json
from django.conf import settings
from django.core.exceptions import ValidationError
from django import forms
from django.core.validators import RegexValidator
from urllib.parse import urlparse
from taler.util.taler_error_codes import ErrorCode
from http import HTTPStatus

##
# Constant value for the biggest number the bank handles.
# This value is just equal to the biggest number that JavaScript
# can handle (because of the wallet).
# FIXME: also defined in views.py.  Need a common.py to contain
# such definitions ?
UINT64_MAX = (2 ** 64) - 1

##
# Pattern for amounts, plain RegEx.
AMOUNT_REGEX = "^[A-Za-z0-9_-]+:([0-9]+)\.?([0-9]+)?$"


##
# Exception class to be raised when a expected URL parameter
# is not found.
class InvalidSession(ValueError):
    ##
    # Init method.
    #
    # @param self the object itself.
    # @param http_status_code the HTTP response code to return
    #        to the caller (client).
    def __init__(self, http_status_code):
        self.hint = "Landed on a broken session"
        self.http_status_code = http_status_code
        super().__init__()


class InternalServerError(Exception):
    def __init__(self, hint):
        self.hint = hint
        self.http_status_code = HTTPStatus.INTERNAL_SERVER_ERROR
        self.taler_error_code = ErrorCode.INTERNAL_LOGIC_ERROR


##
# Exception class to be raised when a JSON
# object does not respect a specification.
class JSONFieldException(ValueError):

    ##
    # Init method.
    #
    # @param self the object itself.
    # @param error object containing the hint, as created by
    #        the Form API.
    # @param http_status_code the HTTP response code to return
    #        to the caller (client).
    def __init__(self, error, http_status_code):
        for k, errors in error.as_data().items():
            messages = [", ".join(error.messages) for error in errors]
            line = f"{k}: " + "".join(messages)
        super(JSONFieldException, self).__init__(line)
        self.hint = line
        self.http_status_code = http_status_code
        self.taler_error_code = TalerErrorCode.BANK_JSON_INVALID


##
# Exception class to be raised when at least one expected URL
# parameter is either not found or malformed.
class URLParamValidationError(ValueError):
    ##
    # Init method.
    #
    # @param self the object itself.
    # @param error object containing the hint.
    # @param http_status_code the HTTP response code to return
    #        to the caller (client).
    def __init__(self, error, http_status_code):
        self.hint = json.stringify(error.as_json())
        self.http_status_code = http_status_code
        self.taler_error_code = ErrorCode.BANK_PARAMETER_MISSING_OR_INVALID
        super().__init__()


class AuthForm(forms.Form):
    type = forms.CharField(
        validators=[
            RegexValidator("^basic$", message="Only 'basic' method provided for now")
        ]
    )

    data = forms.Field(required=False)


class AuthField(forms.Field):
    ##
    # No need to touch the input.  Dict is good
    # and gets validated by the "validate()" method.
    def to_python(self, value):
        return value

    ##
    # Validate input.
    def validate(self, value):
        af = AuthForm(value)
        if not af.is_valid():
            raise ValidationError(json.dumps(af.errors.as_json()))


##
# Common logic to inherit from all the other validators
class BankValidator:
    def __init__(self, validator, data):
        self.validation_result = validator(data)
        if not self.validation_result.is_valid():
            raise JSONFieldException(
                self.validation_result.errors, HTTPStatus.BAD_REQUEST
            )

    def get(self, name, default=None):
        ret = self.validation_result.cleaned_data.get(name)
        if not ret:
            return default
        return ret


class AddIncomingData(BankValidator):
    def __init__(self, data):
        super(AddIncomingData, self).__init__(self.InnerValidator, data)

    class InnerValidator(forms.Form):
        amount = forms.CharField(
            validators=[
                RegexValidator(
                    AMOUNT_REGEX, message="Format CURRENCY:X[.Y] not respected"
                )
            ]
        )
        subject = forms.CharField()
        credit_account = forms.IntegerField(min_value=1)
        exchange_url = forms.URLField()


##
# Subset of /history and /history-range input.
class HistoryParamsBase(forms.Form):
    cancelled = forms.CharField(
        required=False,
        empty_value="show",
        validators=[
            RegexValidator("^(omit|show)$", message="Only 'omit' or 'show' are valid")
        ],
    )

    ordering = forms.CharField(
        required=False,
        empty_value="descending",
        validators=[
            RegexValidator(
                "^(ascending|descending)$",
                message="Only 'ascending' or 'descending' are valid",
            )
        ],
    )

    direction = forms.CharField(
        validators=[
            RegexValidator(
                "^(debit|credit|both|cancel\+|cancel-)$",
                message="Only: debit/credit/both/cancel+/cancel-",
            )
        ]
    )

    # FIXME: adjust min/max values.
    account_number = forms.IntegerField(required=False)


class HistoryParams(BankValidator):
    def __init__(self, data):
        super(HistoryParams, self).__init__(self.InnerValidator, data)

    class InnerValidator(HistoryParamsBase):
        # FIXME: adjust min/max values.
        delta = forms.IntegerField()
        start = forms.IntegerField(required=False)


class PaytoField(forms.Field):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def to_python(self, value):
        return value

    def validate(self, value):

        # The request misses this, default exchange
        # will be used.  NOTE: experience showed that the
        # "required=False" argument given when init the object
        # does NOT prevent this function from being called!
        if not value:
            return
        wire_uri = urlparse(value)
        if "payto" != wire_uri.scheme:
            raise ValidationError("URL is not 'payto'")


class WithdrawHeadless(BankValidator):
    def __init__(self, data):
        super(WithdrawHeadless, self).__init__(self.InnerValidator, data)

    class InnerValidator(forms.Form):
        amount = forms.CharField(
            validators=[
                RegexValidator(
                    AMOUNT_REGEX, message="Format CURRENCY:X[.Y] not respected"
                )
            ]
        )
        reserve_pub = forms.CharField(required=True)
        exchange_payto_uri = PaytoField(required=True)


class WithdrawHeadlessUri(BankValidator):
    def __init__(self, data):
        super(WithdrawHeadlessUri, self).__init__(self.InnerValidator, data)

    class InnerValidator(forms.Form):
        amount = forms.CharField(
            validators=[
                RegexValidator(
                    AMOUNT_REGEX, message="Format CURRENCY:X[.Y] not respected"
                )
            ]
        )



#  This file is part of TALER
#  (C) 2014, 2015, 2016 Taler Systems SA
#
#  TALER is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation; either version 3, or
# (at your option) any later version.
#
#  TALER is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
#  You should have received a copy of the GNU General Public
# License along with TALER; see the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.
#
#  @author Marcello Stanisci
#  @author Florian Dold

from django.urls import include, path
from django.views.generic.base import RedirectView
from django.contrib.auth import views as auth_views
from django.conf.urls.i18n import i18n_patterns
from . import views

# These paths are part of the GNU Taler wire gatweay API
taler_wire_gateway_patterns = [
    path("<str:acct_id>/", views.twg_base, name="twg-base"),
    path("<str:acct_id>/config", views.twg_config, name="twg-config"),
    path(
        "<str:acct_id>/admin/add-incoming",
        views.twg_add_incoming,
        name="twg-add-incoming",
    ),
    path(
        "<str:acct_id>/history/incoming",
        views.twg_history_incoming,
        name="twg-history-incoming",
    ),
    path(
        "<str:acct_id>/history/outgoing",
        views.twg_history_outgoing,
        name="twg-history-outgoing",
    ),
    path("<str:acct_id>/transfer", views.twg_transfer, name="twg-transfer"),
]


taler_bank_integration_api_patterns = [
    path("api/config", views.api_config, name="api-config"),
    path(
        "api/withdrawal-operation/<str:withdraw_id>",
        views.api_withdraw_operation,
        name="api-withdraw-operation",
    ),
]

# These paths are part of the bank access API
taler_bank_access_api_patterns = [
    path(
        "accounts/<str:acct_id>",
        views.bank_accounts_api_balance,
        name="access-api-balance",
    ),
    path(
        "accounts/<str:acct_id>/withdrawals",
        views.bank_accounts_api_create_withdrawal,
        name="access-api-withdrawal",
    ),
    path(
        "accounts/<str:acct_id>/withdrawals/<str:wid>",
        views.bank_accounts_api_get_withdrawal,
        name="access-api-withdrawal-status",
    ),
    path(
        "accounts/<str:acct_id>/withdrawals/<str:wid>/confirm",
        views.bank_accounts_api_confirm_withdrawal,
        name="access-api-withdrawal-confirm",
    ),
    path(
        "accounts/<str:acct_id>/withdrawals/<str:wid>/abort",
        views.bank_accounts_api_abort_withdrawal,
        name="access-api-withdrawal-abort",
    ),
    path("testing/withdraw", views.withdraw_headless, name="testing-withdraw"),
    path("testing/register", views.register_headless, name="testing-withdraw-register"),
]

urlpatterns = [
    path("", include(taler_bank_integration_api_patterns)),
    path("", include(taler_bank_access_api_patterns)),
    path("taler-wire-gateway/", include(taler_wire_gateway_patterns)),
    path(
        "abort-withdrawal/<str:withdraw_id>",
        views.abort_withdrawal,
        name="abort-withdrawal",
    ),
    path("favicon.ico", views.ignore),
    path("config", views.config_view, name="config"),
    path("history", views.serve_history, name="history"),
    path("payto-transfer", views.payto_transfer, name="payto-transfer"),
]

urlpatterns += i18n_patterns(
    path(
        "confirm-withdrawal/<str:withdraw_id>",
        views.confirm_withdrawal,
        name="withdraw-confirm",
    ),
    path("start-withdrawal", views.start_withdrawal, name="start-withdrawal"),
    path(
        "show-withdrawal/<str:withdraw_id>", views.show_withdrawal, name="withdraw-show"
    ),
    path("", RedirectView.as_view(pattern_name="profile"), name="index"),
    path("logout", views.logout_view, name="logout"),
    path("register", views.register, name="register"),
    path("profile", views.profile_page, name="profile"),
    path("payto-form", views.payto_form, name="payto-form"),
    path("wiretransfer-form", views.wiretransfer_form, name="wiretransfer-form"),
    path(
        "login",
        auth_views.LoginView.as_view(
            template_name="login.html",
            authentication_form=views.TalerAuthenticationForm,
        ),
        name="login",
    ),
    path(
        "public-accounts",
        views.serve_public_accounts,
        name="public-accounts",
    ),
    path(
        "public-accounts/<str:name>",
        views.serve_public_accounts,
        name="public-accounts",
    ),
    path(
        "public-accounts/<str:name>/<int:page>",
        views.serve_public_accounts,
        name="public-accounts",
    ),
)



# Generated by Django 3.0.2 on 2020-01-21 15:47

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import talerbank.app.models
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="BankAccount",
            fields=[
                ("is_public", models.BooleanField(default=False)),
                ("account_no", models.AutoField(primary_key=True, serialize=False)),
                (
                    "balance",
                    talerbank.app.models.SignedAmountField(
                        default=talerbank.app.models.get_zero_signed_amount
                    ),
                ),
                (
                    "user",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="TalerWithdrawOperation",
            fields=[
                (
                    "withdraw_id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("amount", talerbank.app.models.AmountField(default=False)),
                ("selection_done", models.BooleanField(default=False)),
                ("confirmation_done", models.BooleanField(default=False)),
                ("aborted", models.BooleanField(default=False)),
                ("selected_reserve_pub", models.TextField(null=True)),
                (
                    "selected_exchange_account",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="selected_exchange_account",
                        to="app.BankAccount",
                    ),
                ),
                (
                    "withdraw_account",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="withdraw_account",
                        to="app.BankAccount",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="BankTransaction",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("amount", talerbank.app.models.AmountField(default=False)),
                (
                    "subject",
                    models.CharField(default="(no subject given)", max_length=200),
                ),
                ("date", models.DateTimeField(auto_now=True, db_index=True)),
                ("cancelled", models.BooleanField(default=False)),
                ("request_uid", models.CharField(max_length=128, unique=True)),
                (
                    "credit_account",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="credit_account",
                        to="app.BankAccount",
                    ),
                ),
                (
                    "debit_account",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="debit_account",
                        to="app.BankAccount",
                    ),
                ),
            ],
        ),
    ]



##
# This file is part of TALER
# (C) 2014, 2015, 2106 Taler Systems SA
#
# TALER is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 3,
# or (at your option) any later version.
#
# TALER is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with TALER; see the file COPYING.  If not,
# see <http://www.gnu.org/licenses/>
#
# @author Marcello Stanisci
# @author Florian Dold
# @brief Create the basic accounts to make the demo bank work.

import sys
import logging
from django.contrib.auth.models import User
from django.db.utils import ProgrammingError, OperationalError
from django.core.management.base import BaseCommand
from django.conf import settings
from ...models import BankAccount
from ...views import wire_transfer
from taler.util.amount import Amount
import getpass
import uuid

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)


class Command(BaseCommand):
    help = "Add money to a bank account (and debit the bank's account)."

    def add_arguments(self, parser):
        parser.add_argument(
            "user",
            type=str,
            metavar="USERNAME",
            help="User that is getting credited with the top-up",
        )
        parser.add_argument(
            "amount",
            type=str,
            metavar="AMOUNT",
            help="Wire transfer's amount, given in the " "CURRENCY:X.Y form.",
        )

    ##
    # Django-specific definition to invoke the account creator
    # @a make_account; it iterates over the list of basic accounts
    # (defined in the settings) and invoke the account creator
    # for each one of them.
    def handle(self, *args, **options):
        user = User.objects.get(username=options["user"])
        # Take the money from the bank's account
        try:
            debit_account = BankAccount.objects.get(
                account_no=1,
            )
        except BankAccount.DoesNotExist:
            LOGGER.error("Debit account (bank's own account) does not exist.")
            sys.exit(1)
        try:
            amount = Amount.parse(options["amount"])
        except BadFormatAmount:
            LOGGER.error("Amount's format is wrong: respect C:X.Y.")
            sys.exit(1)
        try:
            transaction = wire_transfer(
                amount, debit_account, user.bankaccount, "manual top-up by admin"
            )
            print("Transaction id: " + str(transaction.id))
        except Exception as exc:
            LOGGER.error(exc)
            sys.exit(1)



import traceback
import logging
import zlib
from . import urls
from django.http import JsonResponse
from django.urls import reverse
from django.shortcuts import redirect
from .models import BankAccount, BankTransaction
from .views import (
    DebitLimitException,
    SameAccountException,
    LoginFailed,
    UnhandledException,
    set_session_hint,
)

from .schemas import JSONFieldException, URLParamValidationError, InvalidSession

from taler.util.amount import CurrencyMismatchError, AmountFormatError
from taler.util.taler_error_codes import ErrorCode
from http import HTTPStatus

LOGGER = logging.getLogger()


##
# Class decompressing requests.
class DecompressionMiddleware:

    ##
    # Init constructor.
    #
    # @param self the object itself.
    # @param get_response a Django-provided callable that calls
    #        whatever comes next in the chain: a further middleware
    #        or the view itself (please refer to the official
    #        documentation for more details).
    def __init__(self, get_response):
        self.get_response = get_response

    ##
    # This function is transparently invoked by Django when
    # a request traverses the chain made of middleware classes
    # and the view itself as the last element in the chain.
    #
    # Here happens the decompression.
    #
    # @param self this class.
    # @param request Django-specific request object (of the same
    #        type that is handed to views).
    # @return Django-specific response object.
    def __call__(self, request):
        if "deflate" == request.META.get("HTTP_CONTENT_ENCODING"):
            request._body = zlib.decompress(request.body)

        return self.get_response(request)


class ExceptionMiddleware:
    """
    Middleware for handling exceptions not caught directly
    by the application logic.
    """

    def __init__(self, get_response):
        """
        # Init constructor.
        #
        # @param self the object itself.
        # @param get_response a Django-provided callable that calls
        #        whatever comes next in the chain: a further middleware
        #        or the view itself (please refer to the official
        #        documentation for more details).
        """
        self.get_response = get_response

        # Map between endpoints and Web pages to render
        # after the exception gets managed.
        self.render = {
            reverse("profile", urlconf=urls): "profile",
            reverse("register", urlconf=urls): "index",
            reverse("public-accounts", urlconf=urls): "index",
        }

    def __call__(self, request):
        """
        This function is transparently invoked by Django when
        a request traverses the chain made of middleware classes
        and the view itself as the last element in the chain.
        """
        return self.get_response(request)

    def process_exception(self, request, exception):
        """
        Main logic for processing the exception.  It checks
        if the exception captured can be managed, and does it
        if so.  Otherwise, it lets the native handler operate.
        """
        LOGGER.error(f"Error: {exception}, while serving {request.get_full_path()}")

        if hasattr(exception, "taler_error_code"):
            render_to = self.render.get(request.path)
            if not render_to:
                response = JsonResponse(
                    dict(code=exception.taler_error_code.value, error=exception.hint),
                    status=exception.http_status_code,
                )
                response["Access-Control-Allow-Origin"] = "*"
                return response
            set_session_hint(request, success=False, hint=exception.hint)
            return redirect(render_to)
        else:
            LOGGER.error(f"Stack trace: {traceback.format_exc()}")
            return JsonResponse(
                dict(
                    code=ErrorCode.BANK_UNMANAGED_EXCEPTION,
                    hint="unexpected exception",
                    exception=str(exception),
                ),
                status=HTTPStatus.INTERNAL_SERVER_ERROR,
            )
